import datajoint as dj

schema = dj.schema('ibl_dj_alyxraw')


@schema
class AlyxRaw(dj.Manual):
    definition = '''
    uuid: uuid  # pk field (uuid string repr)
    ---
    model: varchar(255)  # alyx 'model'
    '''

    class Field(dj.Part):
        definition = '''
        -> master
        fname: varchar(255)  # field name
        value_idx: tinyint
        ---
        fvalue=null: varchar(10000)  # field value in the position of value_idx
        '''



'''

ibl.ingest

This package contains 'shadowed' copies of main tables for external data load
most classes here will be defined using
<downstream_module>.RealClass.definition;

Only if some merging/disambiguation will definitions be augmented/modified
locally in some way with additional attributes and/or tables to facillitate the
difference. These differences should still result in tables compatbile with
data copying via insert from select (e.g: Foo.insert(Bar.fetch()))

NOTE:

Since downstream modules involve cross-module definitions, those modules should
be imported as 'ds_module' in order to prevent the possibility of accidental
linkages to downstream tables in the upstream schema.

For example, in the scenario:

  - foo.py defines Foo
  - bar.py defines Bar referencing foo.Foo

  - ingest.bar imports .. foo (for some other reason than foo.Foo)
  - ingest.bar imports .. bar (to get foo.Foo schema string)
  - ingest.bar.Bar.definition = bar.Bar.definition

Setting ingest.bar.Bar.definition = bar.Bar.definition creates an accidental
link to downstream foo.Foo table because 'bar' points to the downstream
module. If foo/bar had been imported as ds_foo/ds_bar instead, the table
definition syntax would not properly resolve any 'foo' in the scope of
ingest.bar and the definition would fail, also failing to create the bad link.



In this scheme, the 'correct' implementation would instead be:

  - foo.py defines Foo
  - bar.py defines Bar referencing foo.Foo

  - ingest.bar imports .. foo as ds_foo (for some other reason than foo.Foo)
  - ingest.bar imports .. bar as ds_bar (to get foo.Foo schema string)
  - ingest.bar imports . foo (to get ingest.foo.Foo)
  - ingest.bar.Bar.definition = bar.Bar.definition

Now, ingest.bar.Bar is able to use bar.Bar.definition, but the definition
of ingest.bar.Bar is resolved within the scope of ingest.bar as pointing to
ingest.foo.Foo, creating the proper link to the ingest related table.

While this should not happen in the current architecture, following the pattern
outlined here should prevent it in general and so is a good 'safe practice' to
use for the ingest modules.
'''
import logging
import datajoint as dj
from . import alyxraw


log = logging.getLogger(__name__)


def get_raw_field(key, field, multiple_entries=False, model=None):
    if model:
        query = alyxraw.AlyxRaw.Field & \
            (alyxraw.AlyxRaw & 'model="{}"'.format(model)) & \
            key & 'fname="{}"'.format(field)
    else:
        query = alyxraw.AlyxRaw.Field & key & 'fname="{}"'.format(field)

    return query.fetch('fvalue') \
        if multiple_entries else query.fetch1('fvalue')


class InsertBuffer(object):
    '''
    InsertBuffer: a utility class to help managed chunked inserts
    Currently requires records do not have prerequisites.
    '''
    def __init__(self, rel):
        self._rel = rel
        self._queue = []

    def insert1(self, r):
        self._queue.append(r)

    def insert(self, recs):
        self._queue += recs

    def flush(self, replace=False, skip_duplicates=False,
              ignore_extra_fields=False, chunksz=1):
        '''
        flush the buffer
        XXX: use kwargs?
        XXX: ignore_extra_fields na, requires .insert() support
        '''
        qlen = len(self._queue)
        if qlen > 0 and qlen % chunksz == 0:
            try:
                self._rel.insert(self._queue, skip_duplicates=skip_duplicates,
                                 ignore_extra_fields=ignore_extra_fields)
                self._queue.clear()
                return qlen
            except dj.DataJointError as e:
                log.error('error in flush: {}'.format(e))
                raise
        else:
            return 0



import datajoint as dj
import uuid

from . import alyxraw
from . import get_raw_field as grf


schema = dj.schema(dj.config.get('database.prefix', '') +
                   'ibl_dj_ingest_reference')


@schema
class Lab(dj.Computed):
    # <class 'misc.models.Lab'>
    definition = """
    (lab_uuid) -> alyxraw.AlyxRaw
    ---
    lab_name:               varchar(255)  # name of lab
    institution:            varchar(255)
    address:                varchar(255)
    time_zone:              varchar(255)
    reference_weight_pct:   float
    zscore_weight_pct:      float
    """
    key_source = (alyxraw.AlyxRaw & 'model = "misc.lab"').proj(lab_uuid='uuid')

    def make(self, key):
        key_lab = key.copy()
        key['uuid'] = key['lab_uuid']
        key_lab['lab_name'] = grf(key, 'name')
        key_lab['institution'] = grf(key, 'institution')
        key_lab['address'] = grf(key, 'address')
        key_lab['time_zone'] = grf(key, 'timezone')
        key_lab['reference_weight_pct'] = grf(key, 'reference_weight_pct')
        key_lab['zscore_weight_pct'] = grf(key, 'zscore_weight_pct')
        self.insert1(key_lab)


@schema
class LabMember(dj.Computed):
    # <class 'misc.models.OrderedUser'>
    # <class 'django.contrib.auth.models.User'>
    definition = """
    (user_uuid) -> alyxraw.AlyxRaw
    ---
    user_name:		        varchar(255)	# username
    password:		        varchar(255)	# password
    email=null:		        varchar(255)	# email address
    last_login=null:	    datetime	    # last login
    first_name=null:        varchar(255)	# first name
    last_name=null:		    varchar(255)	# last name
    date_joined:	        datetime	    # date joined
    is_active:		        boolean		    # active
    is_staff:		        boolean		    # staff status
    is_superuser:	        boolean		    # superuser status
    is_stock_manager:       boolean         # stock manager status
    """
    key_source = (alyxraw.AlyxRaw & 'model = "misc.labmember"').proj(
        user_uuid='uuid')

    def make(self, key):
        key_lab_member = key.copy()
        key['uuid'] = key['user_uuid']
        key_lab_member['user_name'] = grf(key, 'username')
        key_lab_member['password'] = grf(key, 'password')
        key_lab_member['email'] = grf(key, 'email')

        last_login = grf(key, 'last_login')
        if last_login != 'None':
            key_lab_member['last_login'] = last_login

        first_name = grf(key, 'first_name')
        if first_name != 'None':
            key_lab_member['first_name'] = first_name

        last_name = grf(key, 'last_name')
        if last_name != 'None':
            key_lab_member['last_name'] = last_name

        key_lab_member['date_joined'] = grf(key, 'date_joined')

        is_active = grf(key, 'is_active')
        key_lab_member['is_active'] = is_active == 'True'

        is_staff = grf(key, 'is_staff')
        key_lab_member['is_staff'] = is_staff == 'True'

        is_superuser = grf(key, 'is_superuser')
        key_lab_member['is_superuser'] = is_superuser == 'True'

        is_stock_manager = grf(key, 'is_stock_manager')
        key_lab_member['is_stock_manager'] = is_stock_manager == 'True'

        self.insert1(key_lab_member)


@schema
class LabMembership(dj.Computed):
    definition = """
    (lab_membership_uuid) -> alyxraw.AlyxRaw
    ---
    lab_name:               varchar(255)
    user_name:              varchar(255)
    role=null:              varchar(255)
    mem_start_date=null:    date
    mem_end_date=null:      date
    """
    key_source = (alyxraw.AlyxRaw & 'model="misc.labmembership"').proj(
        lab_membership_uuid='uuid')

    def make(self, key):
        key_mem = key.copy()
        key['uuid'] = key['lab_membership_uuid']

        lab_uuid = grf(key, 'lab')
        key_mem['lab_name'] = \
            (Lab & dict(lab_uuid=uuid.UUID(lab_uuid))).fetch1('lab_name')

        user_uuid = grf(key, 'user')
        key_mem['user_name'] = \
            (LabMember & dict(user_uuid=uuid.UUID(user_uuid))).fetch1(
                'user_name')

        role = grf(key, 'role')
        if role != 'None':
            key_mem['role'] = role

        start_date = grf(key, 'start_date')
        if start_date != 'None':
            key_mem['mem_start_date'] = start_date

        end_date = grf(key, 'end_date')
        if end_date != 'None':
            key_mem['mem_end_date'] = end_date

        self.insert1(key_mem)


@schema
class LabLocation(dj.Computed):
    definition = """
    (location_uuid) -> alyxraw.AlyxRaw
    ---
    lab_name:           varchar(64)
    location_name:      varchar(255)    # name of the location
    """
    key_source = (alyxraw.AlyxRaw & 'model = "misc.lablocation"').proj(
        location_uuid='uuid')

    def make(self, key):
        key_loc = key.copy()
        key['uuid'] = key['location_uuid']
        key_loc['location_name'] = grf(key, 'name')
        lab_uuid = grf(key, 'lab')
        if lab_uuid == 'None':
            key_loc['lab_name'] = 'cortexlab'
        else:
            key_loc['lab_name'] = \
                (Lab & dict(lab_uuid=uuid.UUID(lab_uuid))).fetch1(
                    'lab_name')

        self.insert1(key_loc)


@schema
class Project(dj.Computed):
    definition = """
    (project_uuid) -> alyxraw.AlyxRaw
    ---
    project_name:                varchar(255)
    project_description=null:    varchar(1024)
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.project"').proj(
        project_uuid='uuid')

    def make(self, key):
        key_proj = key.copy()
        key['uuid'] = key['project_uuid']

        key_proj['project_name'] = grf(key, 'name')
        key_proj['project_description'] = grf(key, 'description')
        self.insert1(key_proj)


@schema
class ProjectLabMember(dj.Manual):
    definition = """
    project_name:   varchar(255)
    user_name:      varchar(255)
    """



import datajoint as dj
import numpy as np
from os import path
from . import acquisition
from . import reference

schema = dj.schema(dj.config.get('database.prefix', '') + 'ibl_ephys')

@schema
class Ephys(dj.Imported):
    definition = """
    -> acquisition.Session
    ---
    ephys_raw_dir:              varchar(256)     # Path of Raw ephys file: array of size nSamples * nChannels. Channels from all probes are included. NOTE: this is huge, and hardly even used. To allow people to load it, we need to add slice capabilities to ONE
    ephys_timestamps:       longblob     # Timestamps for raw ephys timeseries (seconds)
    ephys_start_time:       float        # (seconds)
    ephys_stop_time:       float        # (seconds)
    ephys_duration:         float        # (seconds)
    ephys_sampling_rate:    float        # samples per second
    """

    def make(self, key):
        datapath = path.join(path.sep,'data', '{subject_id}-{session_start_time}'.format(**key)).replace(':', '_')
        ephys_raw_dir = path.join(datapath,'ephys.raw.npy')
        ephys_timestamps = np.load(path.join(datapath,'ephys.timestamps.npy'))[:, 1]

        key['ephys_raw_dir'] = ephys_raw_dir
        key['ephys_timestamps'] = ephys_timestamps
        key['ephys_start_time'] = ephys_timestamps[0]
        key['ephys_stop_time'] = ephys_timestamps[-1]
        key['ephys_duration'] = key['ephys_stop_time'] - key['ephys_start_time']
        key['ephys_sampling_rate'] = 1 / np.median(np.diff(ephys_timestamps))

        self.insert1(key)


@schema
class ProbeModel(dj.Lookup):
    definition = """
    # Description of a particular model of probe.
    probe_model_name: varchar(128)      # String naming probe model, from probe.description
    ---
    channel_counts: smallint            # number of channels in the probe
    """

    class Channel(dj.Part):
        definition = """
        -> master
        channel_id:         smallint     # id of a channel on the probe
        ---
        channel_x_pos:  float   # x position relative to the tip of the probe (um)
        channel_y_pos:  float   # y position relative to the tip of the probe (um)
        """


@schema
class ProbeSet(dj.Imported):
    definition = """
    -> Ephys
    """

    class Probe(dj.Part):
        definition = """
        -> master
        probe_idx:          tinyint     # probe number in this array
        ---
        -> ProbeModel
        probe_set_raw_filename: varchar(256)      # Name of the raw data file this probe was recorded in
        entry_point_rl:    float
        entry_point_ap:    float
        vertical_angle:    float
        horizontal_angle:  float
        axial_angle:       float
        distance_advanced: float
        """


@schema
class Channel(dj.Imported):
    definition = """
    -> ProbeSet.Probe
    -> ProbeModel.Channel
    ---
    channel_index:          smallint    # position within the data array of recording
    channel_ccf_ap:         float       # anterior posterior CCF coordinate (um)
    channel_ccf_dv:         float       # dorsal ventral CCF coordinate (um)
    channel_ccf_lr:         float       # left right CCF coordinate (um)
    -> reference.BrainLocationAcronym   # acronym of the brain location
    channel_raw_row:        smallint     # Each channel's row in its home file (look up via probes.rawFileName), counting from zero. Note some rows don't have a channel, for example if they were sync pulses
    """
    key_source = Ephys


@schema
class ClusterGroup(dj.Imported):
    definition = """
    -> ProbeSet
    ---
    """

    class Cluster(dj.Part):
        definition = """
        -> master
        cluster_id: smallint
        ---
        cluster_mean_waveform:      longblob      # Mean unfiltered waveform of spikes in this cluster (but for neuropixels data will have been hardware filtered): nClusters*nSamples*nChannels
        cluster_template_waveform:  longblob      # Waveform that was used to detect those spikes in Kilosort, in whitened space (or the most representative such waveform if multiple templates were merged)
        cluster_depth :             float         # Depth of mean cluster waveform on probe (µm). 0 means deepest site, positive means above this.
        cluster_waveform_duration:  float         # trough to peak time (ms)
        cluster_amp:                float         # Mean amplitude of each cluster (µV)
        -> ProbeSet.Probe
        (cluster_peak_channel)  -> Channel(channel_id)
        cluster_phy_annotation:     tinyint       # 0 = noise, 1 = MUA, 2 = Good, 3 = Unsorted, other number indicates manual quality score (from 4 to 100)
        """


@schema
class ClusterSpikes(dj.Imported):
    definition = """
    -> ClusterGroup.Cluster
    ---
    cluster_spike_times:    longblob        # spike times of a particular cluster (seconds)
    cluster_spike_depth:    longblob        # Depth along probe of each spike (µm; computed from waveform center of mass). 0 means deepest site, positive means above this
    cluster_spike_amps:     longblob        # Amplitude of each spike (µV)
    """
    key_source = Ephys


@schema
class LFP(dj.Imported):
    definition = """
    -> ProbeSet
    ---
    lfp_raw:              longblob     # LFP: array of size nSamples * nChannels. Channels from all probes are included
    lfp_timestamps:       longblob     # Timestamps for LFP timeseries in seconds
    lfp_start_time:       float        # (seconds)
    lfp_end_time:         float        # (seconds)
    lfp_duration:         float        # (seconds)
    lfp_sampling_rate:    float        # samples per second
    """



import datajoint as dj

from . import subject
from . import reference, subject, action

schema = dj.schema(dj.config.get('database.prefix', '') + 'ibl_dj_acquisition')

'''
To simplify from Alyx schema, dropping file repositories as core
parts of tables and moving to per-table blobs. A facility for
tracking data provenance could still made avaialble via a table
linked into individual model records as a non-primary attribute.

This is probably not ideal w/r/t Alyx -

one idea would be to merge use external storage with example
field def being:

  aligned_movie :  external  # motion-aligned movie

and also have a separate 'phase 1' data schema, which is then coupled
with phase 2+ for higher level data products.

Also note: TimeScale items missing wrt TimeSeries items, which were:

- PupilTracking.xyd
- PupilTracking.movie
- HeadTracking.xy_theta
- HeadTracking.movie

TimeScale not yet defined

# SKIPPED:
# <class 'data.models.DataRepositoryType'>
# <class 'data.models.DataRepository'>
# <class 'data.models.DatasetType'>
# <class 'data.models.Dataset'>
# <class 'data.models.FileRecord'>
# <class 'data.models.DataCollection'>
# <class 'data.models.Timescale'>
# <class 'data.models.TimeSeries'>
# <class 'data.models.EventSeries'>
# <class 'data.models.IntervalSeries'>
'''


@schema
class Session(dj.Manual):
    # <class 'actions.models.Session'>
    definition = """
    -> subject.Subject
    session_start_time:         datetime	# start time
    ---
    session_uuid:               uuid
    session_number=null:        int     	# number
    session_end_time=null:      datetime	# end time
    -> [nullable] reference.Project
    -> [nullable] reference.LabLocation.proj(session_lab='lab_name', session_location='location_name')
    task_protocol=null:         varchar(255)
    session_type=null:		    varchar(255)	# type
    session_narrative=null:     varchar(1024)
    """


@schema
class ChildSession(dj.Manual):
    definition = """
    -> Session
    ---
    (parent_session_start_time) -> Session(session_start_time)
    """


@schema
class SessionUser(dj.Manual):
    definition = """
    -> Session
    -> reference.LabMember
    """


@schema
class SessionProcedure(dj.Manual):
    definition = """
    -> Session
    -> action.ProcedureType
    """


@schema
class WaterAdministrationSession(dj.Manual):
    definition = """
    -> action.WaterAdministration
    ---
    -> Session
    """



import datajoint as dj

schema = dj.schema(dj.config.get('database.prefix', '') + 'ibl_dj_reference')


@schema
class Lab(dj.Lookup):
    # <class 'misc.models.Lab'>
    definition = """
    lab_name:               varchar(255)  # name of lab
    ---
    lab_uuid:               uuid
    institution:            varchar(255)
    address:                varchar(255)
    time_zone:              varchar(255)
    reference_weight_pct:   float
    zscore_weight_pct:      float
    """


@schema
class LabMember(dj.Manual):
    # <class 'misc.models.LabMember'>
    # <class 'django.contrib.auth.models.User'>
    definition = """
    user_name:		        varchar(255)	# username
    ---
    user_uuid:              uuid
    password:		        varchar(255)	# password
    email=null:		        varchar(255)	# email address
    last_login=null:	    datetime	    # last login
    first_name=null:        varchar(255)	# first name
    last_name=null:		    varchar(255)	# last name
    date_joined:	        datetime	    # date joined
    is_active:		        boolean		    # active
    is_staff:		        boolean		    # staff status
    is_superuser:	        boolean		    # superuser status
    is_stock_manager:       boolean         # stock manager status
    groups=null:            blob            #
    user_permissions=null:   blob            #
    """


@schema
class LabMembership(dj.Manual):
    definition = """
    -> Lab
    -> LabMember
    ---
    lab_membership_uuid:    uuid
    role=null:              varchar(255)
    mem_start_date=null:    date
    mem_end_date=null:      date
    """


@schema
class LabLocation(dj.Manual):
    # <class 'misc.models.LabLocation'>
    definition = """
    # The physical location at which an session is performed or appliances are located.
    # This could be a room, a bench, a rig, etc.
    -> Lab
    location_name:      varchar(255)    # name of the location
    ---
    location_uuid:      uuid
    """


@schema
class Project(dj.Lookup):
    definition = """
    project_name:               varchar(255)
    ---
    project_uuid:               uuid
    project_description=null:   varchar(1024)
    """


@schema
class ProjectLabMember(dj.Manual):
    definition = """
    -> Project
    -> LabMember
    """


@schema
class Severity(dj.Lookup):
    definition = """
    severity:			tinyint			# severity
    ---
    severity_desc:		varchar(32)		# severity desc
    """
    contents = (
        (0, ''),
        (1, 'Sub-threshold'),
        (2, 'Mild'),
        (3, 'Moderate'),
        (4, 'Severe'),
        (5, 'Non-recovery'),
    )


@schema
class BrainLocationAcronym(dj.Lookup):
    definition = """
    acronym:  varchar(32) # acronym of a brain location
    ---
    full_name = null: varchar(128) # full name of the brain location
    """
    contents = [
        ['ACA', 'Anterior cingulate area'],
        ['ACB', 'Nucleus accumbens'],
        ['IC', 'Inferior colliculus '],
        ['MOs', 'Secondary motor area'],
        ['MRN', 'Midbrain reticular nucleus'],
        ['root', ''],
        ['RSP', 'Retrosplenial area'],
        ['SCsg', 'Superficial gray layer '],
        ['VISp', 'Primary visual area']
    ]



import datajoint as dj
import json
import uuid

from . import alyxraw, reference
from . import get_raw_field as grf

schema = dj.schema(dj.config.get('database.prefix', '') +
                   'ibl_dj_ingest_subject')

subjects = alyxraw.AlyxRaw & 'model="subjects.subject"'


@schema
class Species(dj.Computed):
    definition = """
    (species_uuid) -> alyxraw.AlyxRaw
    ---
    binomial:           varchar(255)
    species_nickname:   varchar(255)
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.species"').proj(
        species_uuid='uuid')

    def make(self, key):
        key_species = key.copy()
        key['uuid'] = key['species_uuid']
        key_species['binomial'] = grf(key, 'name')
        key_species['species_nickname'] = grf(key, 'nickname')

        self.insert1(key_species)


@schema
class Strain(dj.Computed):
    # <class 'subjects.models.Strain'>
    definition = """
    (strain_uuid) -> alyxraw.AlyxRaw
    ---
    strain_name:		        varchar(255)	# strain name
    strain_description=null:    varchar(255)	# description
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.strain"').proj(
        strain_uuid='uuid')

    def make(self, key):
        key_strain = key.copy()
        key['uuid'] = key['strain_uuid']
        key_strain['strain_name'] = grf(key, 'name')

        description = grf(key, 'description')
        if description != 'None':
            key_strain['strain_description'] = description

        self.insert1(key_strain)


@schema
class Source(dj.Computed):
    # <class 'subjects.models.Source'>
    definition = """
    (source_uuid) -> alyxraw.AlyxRaw
    ---
    source_name:				varchar(255)	# name of source
    source_description=null:	varchar(255)	# description
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.source"').proj(
        source_uuid='uuid')

    def make(self, key):
        key_animal_source = key.copy()
        key['uuid'] = key['source_uuid']
        key_animal_source['source_name'] = grf(key, 'name')

        description = grf(key, 'description')
        if description != 'None':
            key_animal_source['source_description'] = description
        self.insert1(key_animal_source)


@schema
class Sequence(dj.Computed):
    # <class 'subjects.models.Sequence'>
    definition = """
    (sequence_uuid) -> alyxraw.AlyxRaw
    ---
    sequence_name:		        varchar(255)	# informal name
    base_pairs=null:	        varchar(1024)	# base pairs
    sequence_description=null:	varchar(255)	# description
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.sequence"').proj(
        sequence_uuid='uuid')

    def make(self, key):
        key_seq = key.copy()
        key['uuid'] = key['sequence_uuid']
        key_seq['sequence_name'] = grf(key, 'name')

        base_pairs = grf(key, 'base_pairs')
        if base_pairs != 'None':
            key_seq['base_pairs'] = base_pairs

        description = grf(key, 'description')
        if description != 'None':
            key_seq['sequence_description'] = description

        self.insert1(key_seq)


@schema
class Allele(dj.Computed):
    # <class 'subjects.models.Allele'>
    definition = """
    (allele_uuid) -> alyxraw.AlyxRaw
    ---
    allele_name:			    varchar(255)    # informal name
    standard_name=null:		    varchar(255)	# standard name
    allele_source=null:         varchar(255)    # source of the allele
    source_identifier=null:     varchar(255)    # id inside the line provider
    source_url=null:            varchar(255)    # link to the line information
    expression_data_url=null:   varchar(255)    # link to the expression pattern from Allen institute brain atlas
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.allele"').proj(
        allele_uuid='uuid')

    def make(self, key):
        key_allele = key.copy()
        key['uuid'] = key['allele_uuid']
        key_allele['allele_name'] = grf(key, 'nickname')

        standard_name = grf(key, 'name')
        if standard_name != 'None':
            key_allele['standard_name'] = standard_name

        self.insert1(key_allele)


@schema
class AlleleSequence(dj.Manual):
    definition = """
    allele_name:        varchar(255)    # allele name, inherited from Allele
    sequence_name:      varchar(255)    # sequence name, inherited from Sequence
    """


@schema
class Line(dj.Computed):
    # <class 'subjects.models.Line'>
    definition = """
    (line_uuid) -> alyxraw.AlyxRaw
    ---
    binomial:                   varchar(255)	# binomial, inherited from Species
    strain_name=null:           varchar(255)    # strain name, inherited from Strain
    line_name:				    varchar(255)	# line name
    line_description=null:		varchar(2048)	# description
    target_phenotype=null:		varchar(255)	# target phenotype
    line_nickname:				varchar(255)	# auto name
    is_active:				    boolean		    # is active
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.line"').proj(
        line_uuid='uuid')

    def make(self, key):

        key_line = key.copy()
        key['uuid'] = key['line_uuid']

        species_uuid = grf(key, 'species')
        key_line['binomial'] = \
            (Species & dict(species_uuid=uuid.UUID(species_uuid))).fetch1(
                'binomial')

        strain_uuid = grf(key, 'strain')
        if strain_uuid != 'None':
            key_line['strain_name'] = (Strain & dict(
                strain_uuid=uuid.UUID(strain_uuid))).fetch1('strain_name')

        key_line['line_name'] = grf(key, 'name')

        description = grf(key, 'description')
        if description != 'None':
            key_line['line_description'] = description
        key_line['target_phenotype'] = grf(key, 'target_phenotype')
        key_line['line_nickname'] = grf(key, 'nickname')

        active = grf(key, 'is_active')
        key_line['is_active'] = active == "True"

        self.insert1(key_line)


@schema
class LineAllele(dj.Manual):
    definition = """
    line_name:				varchar(255)	# name
    allele_name:			varchar(255)    # informal name
    """


@schema
class Subject(dj.Computed):
    # <class 'subjects.models.Subject'>
    definition = """
    (subject_uuid) -> alyxraw.AlyxRaw
    ---
    subject_nickname:			varchar(255)		# nickname
    sex:			            enum("M", "F", "U")	# sex
    subject_birth_date=null:    date			    # birth date
    subject_line=null:          varchar(255)        # line of the subject
    protocol_number:	        tinyint         	# protocol number
    ear_mark=null:			    varchar(255)		# ear mark
    subject_source=null:        varchar(255)        # source name, inherited from Source
    subject_description=null:   varchar(1024)
    """

    key_source = subjects.proj(subject_uuid='uuid')

    def make(self, key):

        key_subject = key.copy()
        key['uuid'] = key['subject_uuid']

        nickname = grf(key, 'nickname')
        if nickname != 'None':
            key_subject['subject_nickname'] = nickname

        sex = grf(key, 'sex')
        if sex != 'None':
            key_subject['sex'] = sex

        birth_date = grf(key, 'birth_date')
        if birth_date != 'None':
            key_subject['subject_birth_date'] = birth_date

        line_uuid = grf(key, 'line')
        if line_uuid != 'None':
            key_subject['subject_line'] = \
                (Line & dict(line_uuid=uuid.UUID(line_uuid))).fetch1(
                    'line_name')

        key_subject['protocol_number'] = grf(key, 'protocol_number')

        ear_mark = grf(key, 'ear_mark')
        if ear_mark != 'None':
            key_subject['ear_mark'] = ear_mark

        source_uuid = grf(key, 'source')
        if source_uuid != 'None':
            key_subject['subject_source'] = \
                (Source & dict(source_uuid=uuid.UUID(source_uuid))).fetch1(
                    'source_name')

        description = grf(key, 'description')
        if description != 'None':
            key_subject['subject_description'] = description

        self.insert1(key_subject)


@schema
class BreedingPair(dj.Computed):
    # <class 'subjects.models.BreedingPair'>
    definition = """
    (bp_uuid) -> alyxraw.AlyxRaw
    ---
    bp_line=null:           varchar(255)        # line name, inherited from Line
    bp_name:			    varchar(255)		# name of the breeding pair
    bp_description=null:	varchar(2048)		# description
    bp_start_date=null:		date			    # start date
    bp_end_date=null:		date			    # end date
    father=null:            uuid                # subject uuid of dad, inherited from subject
    mother1=null:           uuid                # subject uuid of mom, inherited from subject
    mother2=null:		    uuid                # subject uuid of mom2, if has one, inherited from subject
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.breedingpair"').proj(
        bp_uuid='uuid')

    def make(self, key):
        key_bp = key.copy()
        key['uuid'] = key['bp_uuid']

        line_uuid = grf(key, 'line')
        if line_uuid != 'None':
            key_bp['bp_line'] = \
                (Line & dict(line_uuid=uuid.UUID(line_uuid))).fetch1(
                    'line_name')

        key_bp['bp_name'] = grf(key, 'name')

        description = grf(key, 'description')
        if description != 'None':
            key_bp['bp_description'] = description

        start_date = grf(key, 'start_date')
        if start_date != 'None':
            key_bp['bp_start_date'] = grf(key, 'start_date')

        end_date = grf(key, 'end_date')
        if end_date != 'None':
            key_bp['bp_end_date'] = end_date

        father = grf(key, 'father')
        if father != 'None':
            key_bp['father'] = uuid.UUID(father)

        mother1 = grf(key, 'mother1')
        if mother1 != 'None':
            key_bp['mother1'] = uuid.UUID(mother1)

        mother2 = grf(key, 'mother2')
        if mother2 != 'None':
            key_bp['mother2'] = uuid.UUID(mother2)

        self.insert1(key_bp)


@schema
class Litter(dj.Computed):
    # <class 'subjects.models.Litter'>
    definition = """
    (litter_uuid) -> alyxraw.AlyxRaw
    ---
    litter_name:                    varchar(255)    # name of the litter
    bp_name=null:                   varchar(255)    # name of the breedingpair, inherited from BreedingPair
    litter_line=null:               varchar(255)    # line of the litter
    litter_description=null:        varchar(255)	# description
    litter_birth_date=null:		    date		    # birth date
    """
    key_source = (alyxraw.AlyxRaw & 'model="subjects.litter"').proj(
        litter_uuid='uuid')

    def make(self, key):
        key_litter = key.copy()
        key['uuid'] = key['litter_uuid']

        bp_uuid = grf(key, 'breeding_pair')
        if bp_uuid != 'None':
            key_litter['bp_name'] = \
                (BreedingPair & dict(bp_uuid=uuid.UUID(bp_uuid))).fetch1(
                    'bp_name')

        key_litter['litter_name'] = grf(key, 'name')

        line_uuid = grf(key, 'line')
        if line_uuid != 'None':
            key_litter['litter_line'] = \
                (Line & dict(line_uuid=uuid.UUID(line_uuid))).fetch1(
                    'line_name')

        description = grf(key, 'description')
        if description != 'None':
            key_litter['litter_description'] = description

        birth_date = grf(key, 'birth_date')
        if birth_date != 'None':
            key_litter['litter_birth_date'] = birth_date
        self.insert1(key_litter)


@schema
class LitterSubject(dj.Computed):
    definition = """
    -> Subject
    ---
    litter_name:        varchar(255)
    """

    subjects_with_litter = alyxraw.AlyxRaw.Field & subjects & \
        'fname="litter"' & 'fvalue!="None"'
    key_source = (subjects & subjects_with_litter).proj(
        subject_uuid='uuid')

    def make(self, key):
        key_ls = key.copy()
        key['uuid'] = key['subject_uuid']
        litter = grf(key, 'litter')
        key_ls['litter_name'] = \
            (Litter & dict(litter_uuid=uuid.UUID(litter))).fetch1(
                'litter_name')
        self.insert1(key_ls)


@schema
class SubjectProject(dj.Computed):
    definition = """
    -> Subject
    project_name:           varchar(255)
    """

    subjects_with_projects = alyxraw.AlyxRaw.Field & subjects & \
        'fname="projects"' & 'fvalue!="None"'
    key_source = (subjects & subjects_with_projects).proj(
        subject_uuid='uuid')

    def make(self, key):
        key_s = key.copy()
        key['uuid'] = key['subject_uuid']

        proj_uuids = grf(key, 'projects', multiple_entries=True)
        for proj_uuid in proj_uuids:
            key_sp = key_s.copy()
            key_sp['project_name'] = \
                (reference.Project &
                    dict(project_uuid=uuid.UUID(proj_uuid))).fetch1(
                        'project_name')
            self.insert1(key_sp)


@schema
class SubjectUser(dj.Computed):
    definition = """
    -> Subject
    ---
    responsible_user:       varchar(255)
    """

    subjects_with_user = alyxraw.AlyxRaw.Field & subjects & \
        'fname="responsible_user"' & 'fvalue!="None"'

    key_source = (subjects & subjects_with_user).proj(
        subject_uuid='uuid')

    def make(self, key):
        key_su = key.copy()
        key['uuid'] = key['subject_uuid']

        user = grf(key, 'responsible_user')
        key_su['responsible_user'] = \
            (reference.LabMember &
                dict(user_uuid=uuid.UUID(user))).fetch1('user_name')
        self.insert1(key_su)


@schema
class SubjectLab(dj.Computed):
    definition = """
    -> Subject
    ---
    lab_name:       varchar(255)
    """

    def make(self, key):
        key_sl = key.copy()
        key['uuid'] = key['subject_uuid']
        lab = grf(key, 'lab')
        key_sl['lab_name'] = \
            (reference.Lab &
                dict(lab_uuid=uuid.UUID(lab))).fetch1('lab_name')
        self.insert1(key_sl)


@schema
class Caging(dj.Computed):
    definition = """
    -> Subject
    cage_name:              varchar(255)
    ---
    caging_time=null:       datetime    # time when changed to this cage
    """
    key_source = subjects.proj(subject_uuid='uuid')

    def make(self, key):
        key_cage = key.copy()
        key['uuid'] = key['subject_uuid']

        key_cage['cage_name'] = grf(key, 'cage')
        json_content = grf(key, 'json')
        if json_content != 'None':
            json_dict = json.loads(json_content)
            history = json_dict['history']
            if 'cage' not in history:
                self.insert1(key_cage, skip_duplicates=True)
            else:
                cages = history['cage']
                key_cage_i = key_cage.copy()
                for cage in cages[::-1]:
                    cage_time = cage['date_time']
                    key_cage_i['caging_time'] = cage_time[:-6]
                    self.insert1(key_cage_i, skip_duplicates=True)
                    if cage['value'] != 'None':
                        key_cage_i['cage_name'] = cage['value']


@schema
class UserHistory(dj.Computed):
    definition = """
    -> Subject
    user_name:              varchar(255)  # username
    ---
    user_change_time=null:  datetime      # time when changed to this user
    """

    key_source = subjects.proj(subject_uuid='uuid')

    def make(self, key):
        key_user = key.copy()
        key['uuid'] = key['subject_uuid']

        user = grf(key, 'responsible_user', model='subjects.subject')
        key_user['user_name'] = \
            (reference.LabMember &
             dict(user_uuid=uuid.UUID(user))).fetch1('user_name')

        json_content = grf(key, 'json', model='subjects.subject')
        if json_content != 'None':
            json_dict = json.loads(json_content)
            history = json_dict['history']
            if 'reponsible_user' not in history:
                self.insert1(key_user)
            else:
                users = history['responsible_user']
                key_user_i = key_user.copy()
                for user in users[::-1]:
                    user_change_time = user['date_time']
                    key_user_i['user_change_time'] = user_change_time[:-6]
                    self.insert1(key_user_i)
                    if user['value'] != 'None':
                        user_uuid = user['value']
                        key_user_i['user_name'] = \
                            (reference.LabMember &
                             dict(user_uuid=uuid.UUID(user_uuid))).fetch1(
                                 'user_name')
        else:
            self.insert1(key_user)


@schema
class Weaning(dj.Computed):
    definition = """
    -> Subject
    ---
    wean_date=null:			date			# wean date
    """

    subjects_with_wean = alyxraw.AlyxRaw.Field & subjects & \
        'fname="wean_date"' & 'fvalue!="None"'
    key_source = (subjects & subjects_with_wean).proj(
        subject_uuid='uuid')

    def make(self, key):
        key_weaning = key.copy()
        key['uuid'] = key['subject_uuid']

        key_weaning['wean_date'] = grf(key, 'wean_date')
        self.insert1(key_weaning)


@schema
class Death(dj.Computed):
    definition = """
    -> Subject
    ---
    death_date=null:         date
    """
    subjects_with_death = alyxraw.AlyxRaw.Field & subjects & \
        'fname="death_date"' & 'fvalue!="None"'
    key_source = (subjects & subjects_with_death).proj(
        subject_uuid='uuid')

    def make(self, key):
        key_death = key.copy()
        key['uuid'] = key['subject_uuid']
        key_death['death_date'] = grf(key, 'death_date')
        self.insert1(key_death)


@schema
class GenotypeTest(dj.Computed):
    # <class 'subjects.models.Subject'>
    # <class 'subjects.models.Zygosity'>
    # genotype = models.ManyToManyField('Allele', through='Zygosity')
    definition = """
    (genotype_test_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:       uuid
    sequence_name:      varchar(255)              # inherited from Sequence
    test_result:		enum("Present", "Absent") # test result
    """
    key_source = (alyxraw.AlyxRaw & 'model = "subjects.genotypetest"').proj(
        genotype_test_uuid='uuid')

    def make(self, key):
        key_gt = key.copy()
        key['uuid'] = key['genotype_test_uuid']
        key_gt['subject_uuid'] = uuid.UUID(grf(key, 'subject'))

        sequence_uuid = grf(key, 'sequence')
        key_gt['sequence_name'] = \
            (Sequence & dict(sequence_uuid=uuid.UUID(sequence_uuid))).fetch1(
                'sequence_name')

        test_result = grf(key, 'test_result')
        key_gt['test_result'] = 'Present' if test_result else 'Absent'
        self.insert1(key_gt)


@schema
class Zygosity(dj.Computed):
    # <class 'subjects.models.Subject'>
    # <class 'subjects.models.Zygosity'>
    # genotype = models.ManyToManyField('Allele', through='Zygosity')
    definition = """
    (zygosity_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:       uuid
    allele_name:        varchar(255)            # inherited from Allele
    zygosity:           enum("Present", "Absent", "Homozygous", "Heterozygous") # zygosity
    """
    key_source = (alyxraw.AlyxRaw & 'model = "subjects.zygosity"').proj(
        zygosity_uuid='uuid')

    def make(self, key):

        key_zg = key.copy()
        key['uuid'] = key['zygosity_uuid']
        key_zg['subject_uuid'] = uuid.UUID(grf(key, 'subject'))

        if not len(Subject & key_zg):
            return

        allele_uuid = grf(key, 'allele')
        key_zg['allele_name'] = \
            (Allele & dict(allele_uuid=uuid.UUID(allele_uuid))).fetch1(
                'allele_name')

        zygosity = grf(key, 'zygosity')
        zygosity_types = {
            '0': 'Absent',
            '1': 'Heterozygous',
            '2': 'Homozygous',
            '3': 'Present'
        }
        key_zg['zygosity'] = zygosity_types[zygosity]

        self.insert1(key_zg)


@schema
class Implant(dj.Computed):
    # <class 'subjects.models.Subject'>
    definition = """
    -> Subject
    ---
    implant_weight:		        float			    # implant weight
    adverse_effects=null:	    varchar(2048)		# adverse effects
    actual_severity=null:       tinyint             # actual severity, inherited from Severity
    protocol_number:            tinyint
    """
    subjects_with_implant = alyxraw.AlyxRaw.Field & subjects & \
        'fname="implant_weight"' & 'fvalue!="None"'
    key_source = (subjects & subjects_with_implant).proj(
        subject_uuid='uuid')

    def make(self, key):
        key_implant = key.copy()
        key['uuid'] = key['subject_uuid']

        key_implant['implant_weight'] = float(grf(key, 'implant_weight'))

        adverse_effects = grf(key, 'adverse_effects')
        if adverse_effects != 'None':
            key_implant['adverse_effects'] = adverse_effects

        actual_severity = grf(key, 'actual_severity')
        if actual_severity != 'None':
            key_implant['actual_severity'] = int(actual_severity)

        key_implant['protocol_number'] = int(grf(key, 'protocol_number'))

        self.insert1(key_implant)



import datajoint as dj
from . import reference

schema = dj.schema(dj.config.get('database.prefix', '') + 'ibl_dj_subject')


# Actions:
# Refactor questions w/r/t old objs:
#
# - <class 'actions.models.ProcedureType'>: SKIPPED
#   What does this do aside from provide a description?
#   should be inclued for e.g. steps, etc?
# - <class 'actions.models.BaseAction'>: SKIPPED
#   Other than key info, does this provide anything other than 'narritive'?
#   if so, needed?


@schema
class Species(dj.Lookup):
    # <class 'subjects.models.Species'>
    definition = """
    binomial:			    varchar(255)	# binomial
    ---
    species_uuid:           uuid
    species_nickname:		varchar(255)	# nick name
    """


@schema
class Source(dj.Lookup):
    # <class 'subjects.models.Source'>
    definition = """
    source_name:				varchar(255)	# name of source
    ---
    source_uuid:                uuid
    source_description=null:	varchar(255)	# description
    """


@schema
class Strain(dj.Lookup):
    # <class 'subjects.models.Strain'>
    definition = """
    strain_name:		        varchar(255)	# strain name
    ---
    strain_uuid:                uuid
    strain_description=null:    varchar(255)	# description
    """


@schema
class Sequence(dj.Lookup):
    # <class 'subjects.models.Sequence'>
    definition = """
    sequence_name:		        varchar(255)	# informal name
    ---
    sequence_uuid:              uuid
    base_pairs=null:	        varchar(1024)	# base pairs
    sequence_description=null:	varchar(255)	# description
    """


@schema
class Allele(dj.Lookup):
    # <class 'subjects.models.Allele'>
    definition = """
    allele_name:			    varchar(255)    # informal name
    ---
    allele_uuid:                uuid
    standard_name=null:		    varchar(255)	# standard name
    -> [nullable] Source
    allele_source=null:         varchar(255)    # source of the allele
    source_identifier=null:     varchar(255)    # id inside the line provider
    source_url=null:            varchar(255)    # link to the line information
    expression_data_url=null:   varchar(255)    # link to the expression pattern from Allen institute brain atlas
    """


@schema
class AlleleSequence(dj.Lookup):
    definition = """
    -> Allele
    -> Sequence
    """


@schema
class Line(dj.Lookup):
    # <class 'subjects.models.Line'>
    definition = """
    line_name:				varchar(255)	# name
    ---
    -> Species
    -> [nullable] Strain
    line_uuid:              uuid
    line_description=null:	varchar(2048)	# description
    target_phenotype=null:	varchar(255)	# target phenotype
    line_nickname:		    varchar(255)	# nickname
    is_active:				boolean		    # is active
    """


@schema
class LineAllele(dj.Lookup):
    definition = """
    -> Line
    -> Allele
    """


@schema
class Subject(dj.Manual):
    # <class 'subjects.models.Subject'>
    definition = """
    subject_uuid:               uuid
    ---
    subject_nickname:		    varchar(255)		# nickname
    sex:			            enum("M", "F", "U")	# sex
    subject_birth_date=null:	date			    # birth date
    ear_mark=null:			    varchar(255)		# ear mark
    -> [nullable] Line.proj(subject_line="line_name")
    -> [nullable] Source.proj(subject_source='source_name')
    protocol_number:            tinyint         	# protocol number
    subject_description=null:   varchar(1024)
    """


@schema
class BreedingPair(dj.Manual):
    # <class 'subjects.models.BreedingPair'>
    definition = """
    bp_name:			    varchar(255)		    # name
    ---
    -> [nullable] Line.proj(bp_line='line_name')
    bp_uuid:                uuid
    bp_description=null:	varchar(2048)		    # description
    bp_start_date=null:	    date		            # start date
    bp_end_date=null:		date			        # end date
    -> [nullable] Subject.proj(father='subject_uuid')    # father
    -> [nullable] Subject.proj(mother1='subject_uuid')   # mother1
    -> [nullable] Subject.proj(mother2='subject_uuid')	 # mother2
    """


@schema
class Litter(dj.Manual):
    # <class 'subjects.models.Litter'>
    definition = """
    litter_name:                    varchar(255)
    ---
    -> [nullable] BreedingPair
    -> [nullable] Line.proj(litter_line='line_name')
    litter_uuid:			        uuid	    # litter uuid
    litter_descriptive_name=null:	varchar(255)	# descriptive name
    litter_description=null:	    varchar(255)	# description
    litter_birth_date=null:			date		    # birth date
    """


@schema
class LitterSubject(dj.Manual):
    # litter subject membership table
    definition = """
    -> Subject
    ---
    -> Litter
    """


@schema
class SubjectProject(dj.Manual):
    definition = """
    -> Subject
    -> reference.Project
    """


@schema
class SubjectUser(dj.Manual):
    definition = """
    -> Subject
    ---
    -> reference.LabMember.proj(responsible_user='user_name')
    """


@schema
class SubjectLab(dj.Manual):
    definition = """
    -> Subject
    ---
    -> reference.Lab
    """


@schema
class Caging(dj.Manual):
    # <class 'subjects.models.Subject'>
    definition = """
    -> Subject
    cage_name:              varchar(255)        # cage name
    ---
    caging_time=null:	    datetime			# cage
    """


@schema
class UserHistory(dj.Manual):
    definition = """
    -> Subject
    -> reference.LabMember
    ---
    user_change_time=null:   datetime      # time when changed to this user
    """


@schema
class Weaning(dj.Manual):
    # <class 'subjects.models.Subject'>
    definition = """
    -> Subject
    ---
    wean_date:			date			# wean date
    """


@schema
class GenotypeTest(dj.Manual):
    # <class 'subjects.models.Subject'>
    # <class 'subjects.models.Zygosity'>
    # genotype = models.ManyToManyField('Allele', through='Zygosity')
    definition = """
    -> Subject
    -> Sequence
    genotype_test_uuid:		    uuid     # genotype test id
    ---
    test_result:		        enum("Present", "Absent")		# test result
    """


@schema
class Zygosity(dj.Manual):
    # <class 'subjects.models.Subject'>
    # <class 'subjects.models.Zygosity'>
    # genotype = models.ManyToManyField('Allele', through='Zygosity')
    definition = """
    -> Subject
    -> Allele
    ---
    zygosity_uuid:      uuid
    zygosity:		    enum("Present", "Absent", "Homozygous", "Heterozygous")  # zygosity
    """


@schema
class Implant(dj.Manual):
    # <class 'subjects.models.Subject'>
    definition = """
    -> Subject
    ---
    implant_weight:		    float			    # implant weight
    protocol_number:        tinyint		        # protocol number
    adverse_effects=null:   varchar(2048)		# adverse effects
    (actual_severity)		-> [nullable] reference.Severity   # actual severity
    """


@schema
class Death(dj.Manual):
    # <class 'subjects.models.Subject'>
    definition = """
    -> Subject
    ---
    death_date:                 date                    # death date
    """



import datajoint as dj
from . import reference, subject

schema = dj.schema(dj.config.get('database.prefix', '') + 'ibl_dj_action')


@schema
class ProcedureType(dj.Manual):
    definition = """
    procedure_type_name:                varchar(255)
    ---
    procedure_type_uuid:                uuid
    procedure_type_description=null:    varchar(1024)
    """


@schema
class Weighing(dj.Manual):
    # <class 'actions.models.Weighing'>
    definition = """
    -> subject.Subject
    weighing_time:		datetime		# date time
    ---
    weigh_uuid:        uuid
    weight:			    float			# weight
    -> [nullable] reference.LabMember.proj(weighing_user="user_name")
    """


@schema
class WaterType(dj.Lookup):
    definition = """
    watertype_name:     varchar(255)
    ---
    watertype_uuid:     uuid
    """


@schema
class WaterAdministration(dj.Manual):
    # <class 'actions.models.WaterAdministration'>
    definition = """
    -> subject.Subject
    administration_time:	    datetime		# date time
    ---
    wateradmin_uuid:            uuid
    water_administered=null:    float			# water administered
    adlib:                      boolean
    -> WaterType
    -> [nullable] reference.LabMember.proj(administration_user="user_name")
    """


@schema
class WaterRestriction(dj.Manual):
    # <class 'actions.models.WaterRestriction'>
    definition = """
    -> subject.Subject
    restriction_start_time:     datetime	# start time
    ---
    restriction_uuid:           uuid
    restriction_end_time=null:  datetime	# end time
    reference_weight:           float
    restriction_narrative=null: varchar(1024)
    -> [nullable] reference.LabLocation.proj(restriction_lab='lab_name', restriction_location='location_name')
    """


@schema
class WaterRestrictionUser(dj.Manual):
    definition = """
    -> WaterRestriction
    -> reference.LabMember
    """


@schema
class WaterRestrictionProcedure(dj.Manual):
    definition = """
    -> WaterRestriction
    -> ProcedureType
    """


@schema
class Surgery(dj.Manual):
    # <class 'actions.models.Surgery'>
    definition = """
    -> subject.Subject
    surgery_start_time:		datetime        # surgery start time
    ---
    surgery_uuid:           uuid
    surgery_end_time=null:  datetime        # surgery end time
    -> [nullable] reference.LabLocation.proj(surgery_lab='lab_name', surgery_location='location_name')
    surgery_outcome_type:   enum('None', 'a', 'n', 'r')	    # outcome type
    surgery_narrative=null: varchar(2048)	# narrative
    """


@schema
class SurgeryUser(dj.Manual):
    definition = """
    -> Surgery
    -> reference.LabMember
    """


@schema
class SurgeryProcedure(dj.Manual):
    definition = """
    -> Surgery
    -> ProcedureType
    """


@schema
class VirusInjection(dj.Manual):
    # <class 'actions.models.VirusInjection'>
    # XXX: user was m2m field in django
    definition = """
    -> subject.Subject
    injection_time:		    datetime        # injection time
    ---
    injection_uuid:         uuid
    injection_volume:		float   		# injection volume
    rate_of_injection:		float           # rate of injection
    injection_type:		    varchar(255)    # injection type
    """


@schema
class OtherAction(dj.Manual):
    # <class 'actions.models.OtherAction'>
    definition = """
    -> subject.Subject
    other_action_start_time:    datetime	# start time
    ---
    other_action_uuid:          uuid
    other_action_end_time=null: datetime	# end time
    description=null:           varchar(1024)    # description
    -> [nullable] reference.LabLocation.proj(other_action_lab='lab_name', other_action_location='location_name')
    """


@schema
class OtherActionUser(dj.Manual):
    # <class 'actions.models.OtherAction'>
    definition = """
    -> subject.Subject
    other_action_start_time:    datetime	# start time
    ---
    user_name:          varchar(255)
    """


@schema
class OtherActionProcedure(dj.Manual):
    # <class 'actions.models.OtherAction'>
    definition = """
    -> subject.Subject
    other_action_start_time:    datetime	# start time
    ---
    procedure_type_name:        varchar(255)
    """



#! /usr/bin/env python3

import os
import sys
import logging


from code import interact

# BOOKMARK: ensure loading
from ibl_pipeline import reference
from ibl_pipeline import subject
from ibl_pipeline import acquisition
from ibl_pipeline import behavior
from ibl_pipeline import ephys


log = logging.getLogger(__name__)
__all__ = [reference, subject, acquisition, behavior, ephys]


def usage_exit():
    print("usage: {p} [{c}]"
          .format(p=os.path.basename(sys.argv[0]),
                  c='|'.join(list(actions.keys()))))
    sys.exit(0)


def logsetup(*args):
    logging.basicConfig(level=logging.ERROR)
    log.setLevel(logging.DEBUG)
    logging.getLogger('ibl').setLevel(logging.DEBUG)
    logging.getLogger('ibl.ingest').setLevel(logging.DEBUG)


def shell(*args):
    interact('ibl shell.\n\nschema modules:\n\n  - {m}\n'
             .format(m='\n  - '.join(
                 '.'.join(m.__name__.split('.')[1:]) for m in __all__)),
             local=globals())


def ingest(*args):
    # local import so db is only created created/accessed if/when ingesting
    from ibl_pipeline.ingest import (reference as ingest_reference,
                                    subject as ingest_subject,
                                    acquisition as ingest_acquisition)
    for mod in [ingest_reference, ingest_subject, ingest_acquisition]:
        pass


actions = {
    'shell': shell,
    'ingest': ingest,
}


if __name__ == '__main__':

    if len(sys.argv) < 2 or sys.argv[1] not in actions:
        usage_exit()

    logsetup()

    action = sys.argv[1]
    actions[action](sys.argv[2:])



"""
functions to compare contents in shadow tables and real tables.
"""
import logging
import datajoint as dj

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

formatter = logging.Formatter('%(asctime)s:%(name)s:%(message)s')
file_handler = logging.FileHandler('table-diff.log')
file_handler.setFormatter(formatter)

logger.addHandler(file_handler)


def show(tablepairs, comment=''):
    print('# {} table contents'.format(comment))
    for p in tablepairs:
        print(p)
        for t in tablepairs[p]:
            print(t)
            print(t())


def push(tablenames, tablepairs):
    try:
        print('# pushing data')
        for tname in tablenames:
            source_tbl, dest_tbl = tablepairs[tname]
            dest_tbl.insert(source_tbl() - dest_tbl().proj())
    except Exception as e:
        print('# push error: {e}'.format(e=e))


def diff(tablenames, tablepairs):
    ndiff = 0
    for t in tablenames:
        a, b = tablepairs[t]

        for i in ((a - b.proj()).fetch('KEY')):
            logger.info('# {} only in {} - record deleted?'.format(i, a))

        for i in ((b - a.proj()).fetch('KEY')):
            logger.info('# {} only in {} - record deleted?'.format(i, b))

        common = (a & b.proj(*a.primary_key))
        kstr = ', '.join(a.primary_key)
        srcrecs = (a & common.proj()).fetch(order_by=kstr, as_dict=True)
        dstrecs = (b & common.proj()).fetch(order_by=kstr, as_dict=True)

        for srce, dste in zip(srcrecs, dstrecs):
            for attr in srce:
                srcv = srce[attr]
                dstv = dste[attr]
                if srcv != dstv:
                    print('# {t}.{a}: {s} != {d}'.format(
                        t=t, a=attr, s=srcv, d=dstv))
                    ndiff += 1

        logger.info('# {} total differences.'.format(ndiff))



def drop(schemas):
    for s in schemas:
        s.schema.drop(force=True)


import datajoint as dj
import numpy as np
import pandas as pd
from os import path
import logging
from . import reference, subject, acquisition, data
from .ingest import get_raw_field as grf
try:
    from oneibl.one import ONE
except:
    # TODO: consider issuing a warning about not being able to perform ingestion without ONE access
    pass

logger = logging.getLogger(__name__)
schema = dj.schema(dj.config.get('database.prefix', '') + 'ibl_dj_behavior')


@schema
class Eye(dj.Imported):
    definition = """
    # eye recording
    -> acquisition.Session
    ---
    eye_sample_ids:     longblob        # Sample ids corresponding to the timestamps
    eye_timestamps:     longblob        # Timestamps for pupil tracking timeseries (seconds)
    eye_raw_dir=null:   varchar(256)    # directory of the raw datafile
    eye_area:           longblob        # Area of pupil (pixels^2)
    eye_x_pos:          longblob        # x position of pupil (pixels)
    eye_y_pos:          longblob        # y position of pupil (pixels)
    eye_blink:          longblob        # Boolean array saying whether eye was blinking in each frame
    eye_fps:            float           # Frames per second
    eye_start_time:     float           # (seconds)
    eye_end_time:       float           # (seconds)
    """

    key_source = acquisition.Session & (data.FileRecord & {'exists': 1} & 'dataset_name in \
                    ("_ibl_eye.area.npy", "_ibl_eye.blink.npy", \
                     "_ibl_eye.xyPos.npy", "_ibl_eye.timestamps.npy")')

    def make(self, key):

        eID = str((acquisition.Session & key).fetch1('session_uuid'))
        eye_area, eye_blink, eye_xypos, eye_timestamps = \
            ONE().load(eID, dataset_types=['_ibl_eye.area',
                                           '_ibl_eye.blink',
                                           '_ibl_eye.xypos',
                                           '_ibl_eye.timestamps'])

        eye_sample_ids = eye_timestamps[:, 0]
        eye_timestamps = eye_timestamps[:, 1]

        assert len(np.unique(np.array([len(eye_xypos),
                                       len(eye_blink),
                                       len(eye_area)]))) == 1, \
            'Loaded eye files do not have the same length'

        key['eye_sample_ids'] = eye_sample_ids
        key['eye_timestamps'] = eye_timestamps
        key['eye_area'] = eye_area
        key['eye_x_pos'] = eye_xypos[:, 0]
        key['eye_y_pos'] = eye_xypos[:, 1]
        key['eye_blink'] = eye_blink
        key['eye_fps'] = 1 / np.median(np.diff(eye_timestamps))
        key['eye_start_time'] = eye_timestamps[0]
        key['eye_end_time'] = eye_timestamps[-1]

        self.insert1(key)
        logger.info('Populated an Eye tuple for subject {subject_uuid} \
            on session started at {session_start_time}'.format(
            **key))


@schema
class CompleteWheelSession(dj.Computed):
    definition = """
    # sessions that are complete with wheel related information and thus may be ingested
    -> acquisition.Session
    ---
    wheel_session_complete: bool              # whether the session is complete
    """

    required_datasets = ["_ibl_wheel.position.npy",
                         "_ibl_wheel.velocity.npy",
                         "_ibl_wheel.timestamps.npy"]

    def make(self, key):
        datasets = (data.FileRecord & key & {'exists': 1}).fetch(
            'dataset_name')
        key['wheel_session_complete'] = bool(
            np.all([req_ds in datasets for req_ds in self.required_datasets]))
        self.insert1(key)


@schema
class Wheel(dj.Imported):
    definition = """
    # raw wheel recording
    -> acquisition.Session
    ---
    wheel_position:         longblob  # Absolute position of wheel (cm)
    wheel_velocity:         longblob  # Signed velocity of wheel (cm/s) positive = CW
    wheel_sample_ids:       longblob  # Sample ids corresponding to the timestamps
    wheel_timestamps:       longblob  # Timestamps for wheel timeseries (seconds)
    wheel_start_time:       float     # Start time of wheel recording (seconds)
    wheel_end_time:         float     # End time of wheel recording (seconds)
    wheel_duration:         float     # Duration time of wheel recording (seconds)
    wheel_sampling_rate:    float     # Samples per second
    """

    key_source = CompleteWheelSession & 'wheel_session_complete = 1'

    def make(self, key):

        eID = str((acquisition.Session & key).fetch1('session_uuid'))
        wheel_position, wheel_velocity, wheel_timestamps = \
            ONE().load(eID, dataset_types=['_ibl_wheel.position',
                                           '_ibl_wheel.velocity',
                                           '_ibl_wheel.timestamps'])

        wheel_sample_ids = wheel_timestamps[:, 0]
        wheel_timestamps = wheel_timestamps[:, 1]
        wheel_sampling_rate = 1 / np.median(np.diff(wheel_timestamps))

        key['wheel_position'] = wheel_position
        key['wheel_velocity'] = wheel_velocity
        key['wheel_sample_ids'] = wheel_sample_ids
        key['wheel_timestamps'] = wheel_timestamps
        key['wheel_start_time'] = wheel_timestamps[0]
        key['wheel_end_time'] = wheel_timestamps[-1]
        key['wheel_duration'] = wheel_timestamps[-1] - wheel_timestamps[0]
        key['wheel_sampling_rate'] = wheel_sampling_rate

        self.insert1(key)
        logger.info('Populated a Wheel tuple for subject {subject_uuid} \
            in session on {session_start_time}'.format(**key))


@schema
class WheelMoveType(dj.Lookup):
    definition = """
    wheel_move_type:   varchar(64)   # movement type
    """
    contents = [['CW'], ['CCW'], ['Flinch'], ['Other']]


@schema
class CompleteWheelMoveSession(dj.Computed):
    definition = """
    # sessions that are complete with wheel related information and thus may be ingested
    -> acquisition.Session
    ---
    wheelmove_session_complete: bool  # whether the session is complete
    """

    required_datasets = ["_ibl_wheelMoves.intervals.npy",
                         "_ibl_wheelMoves.type.csv"]

    def make(self, key):
        datasets = (data.FileRecord & key & {'exists': 1}).fetch(
            'dataset_name')
        key['wheelmove_session_complete'] = bool(
            np.all([req_ds in datasets for req_ds in self.required_datasets]))
        self.insert1(key)


@schema
class WheelMoveSet(dj.Imported):
    definition = """
    # detected wheel movements
    -> acquisition.Session
    ---
    wheel_move_number : int     # total number of movements in this set
    """
    key_source = CompleteWheelMoveSession & 'wheelmove_session_complete = 1'

    def make(self, key):
        wheel_move_key = key.copy()

        eID = str((acquisition.Session & key).fetch1('session_uuid'))
        wheel_moves_intervals, wheel_moves_types = \
            ONE().load(eID, dataset_types=['_ibl_wheelMoves.intervals',
                                           '_ibl_wheelMoves.type'])

        wheel_moves_types = wheel_moves_types.columns

        assert len(np.unique(np.array([len(wheel_moves_intervals),
                                       len(wheel_moves_types)]))) == 1, \
            'Loaded wheel move files do not have the same length'

        key['wheel_move_number'] = len(wheel_moves_types)
        self.insert1(key)

        for idx_move in range(len(wheel_moves_types)):
            wheel_move_key['wheel_move_id'] = idx_move + 1
            wheel_move_key['wheel_move_start_time'] = \
                wheel_moves_intervals[idx_move, 0]
            wheel_move_key['wheel_move_end_time'] = \
                wheel_moves_intervals[idx_move, 1]

            wheel_move_type = wheel_moves_types[idx_move]
            if 'CCW' in wheel_move_type:
                wheel_move_key['wheel_move_type'] = 'CCW'
            elif 'flinch' in wheel_move_type:
                wheel_move_key['wheel_move_type'] = 'flinch'
            else:
                wheel_move_key['wheel_move_type'] = 'CW'

            self.WheelMove().insert1(wheel_move_key)

        logger.info('Populated a WheelMoveSet and all WheelMove tuples for \
            subject {subject_uuid} in session started at \
            {session_start_time}'.format(**key))

    class WheelMove(dj.Part):
        definition = """
        -> master
        wheel_move_id:          int     # identifier of a wheel movement
        ---
        wheel_move_start_time:  float   # onset time of the detected wheel movement (seconds)
        wheel_move_end_time:    float   # offset time of the detected wheel movement (seconds)
        -> WheelMoveType
        """


@schema
class SparseNoise(dj.Imported):
    definition = """
    -> acquisition.Session
    ---
    sparse_noise_x_pos:  longblob	# x coordiate on screen of sparse noise stimulus squares (WHAT UNIT?)
    sparse_noise_y_pos:  longblob	# y coordiate on screen of sparse noise stimulus squares (WHAT UNIT?)
    sparse_noise_times:  longblob	# times of those stimulus squares appeared in universal seconds
    """

    key_source = acquisition.Session & \
        (data.FileRecord & {'exists': 1} &
         'dataset_name in ("_ibl_sparseNoise.positions.npy", \
                           "_ibl_sparseNoise.times.npy")')

    def make(self, key):

        eID = str((acquisition.Session & key).fetch1('session_uuid'))

        sparse_noise_positions, sparse_noise_times = \
            ONE().load(eID, dataset_types=['_ibl_sparseNoise.positions',
                                           '_ns_sparseNoise.times'])

        assert len(np.unique(np.array([len(sparse_noise_positions),
                                       len(sparse_noise_times)]))) == 1, \
            'Loaded sparse noise files do not have the same length'

        key['sparse_noise_x_pos'] = sparse_noise_positions[:, 0],
        key['sparse_noise_y_pos'] = sparse_noise_positions[:, 1],
        key['sparse_noise_times'] = sparse_noise_times
        self.insert1(key)
        logger.info('Populated a SparseNoise tuple for subject {subject_uuid} \
            in session started at {session_start_time}'.format(**key))


@schema
class ExtraRewards(dj.Imported):
    definition = """
    # information about extra rewards given manually by the experimenter
    -> acquisition.Session
    ---
    extra_rewards_times: longblob 			# times of extra rewards (seconds)
    """

    key_source = acquisition.Session & \
        (data.FileRecord & {'exists': 1} &
            'dataset_name in ("_ibl_extraRewards.times.npy")')

    def make(self, key):

        eID = (acquisition.Session & key).fetch1('session_uuid')

        extra_rewards_times = \
            ONE().load(eID, dataset_types=['_ibl_extraRewards.times'])

        key['extra_rewards_times'] = extra_rewards_times

        self.insert1(key)

        logger.info('Populated an ExtraRewards tuple for \
            subject {subject_uuid} in session started at \
            {session_start_time}'.format(**key))


@schema
class SpontaneousTimeSet(dj.Imported):
    definition = """
    # times when no other protocol was going on for at least 30 sec or so
    -> acquisition.Session
    ---
    spontaneous_time_total_num:   int   # total number of the spontaneous time periods
    """

    key_source = acquisition.Session & \
        (data.FileRecord & {'exists': 1} &
         'dataset_name in ("_ibl_spontaneous.intervals.npy")')

    def make(self, key):
        spon_time_key = key.copy()

        eID = str((acquisition.Session & key).fetch1('session_uuid'))

        spontaneous_intervals = \
            ONE().load(eID, dataset_types=['_ibl_spontaneous.intervals'])

        key['spontaneous_time_total_num'] = len(spontaneous_intervals)
        self.insert1(key)

        for idx_spon in range(len(spontaneous_intervals)):
            spon_time_key['spontaneous_time_id'] = idx_spon + 1
            spon_time_key['spontaneous_start_time'] = \
                spontaneous_intervals[idx_spon, 0]
            spon_time_key['spontaneous_end_time'] = \
                spontaneous_intervals[idx_spon, 1]
            spon_time_key['spontaneous_time_duration'] = \
                float(np.diff(spontaneous_intervals[idx_spon, :]))
            self.SpontaneousTime().insert1(spon_time_key)

        logger.info('Populated a SpontaneousTimeSet tuple and all \
            Spontaneoustime tuples for subject {subject_uuid} in \
                session started at {session_start_time}'.format(**key))

    class SpontaneousTime(dj.Part):
        definition = """
        -> master
        spontaneous_time_id: int    # index of spontanoeous time
        ---
        spontaneous_start_time:     float    # (seconds)
        spontaneous_end_time:       float    # (seconds)
        spontaneous_time_duration:  float    # (seconds)
        """


@schema
class Lick(dj.Imported):
    definition = """
    # detected licks
    -> acquisition.Session
    ---
    lick_times:             longblob  # Times of licks
    lick_piezo_raw = null:  longblob  # Raw lick trace (volts)
    lick_sample_ids:        longblob  # Sample ids corresponding to the timestamps
    lick_piezo_timestamps:  longblob  # Timestamps for lick trace timeseries (seconds)
    lick_start_time:        float     # recording start time (seconds)
    lick_end_time:          float     # recording end time (seconds)]
    lick_sampling_rate:     float     # number of samples per second
    """

    key_source = acquisition.Session & \
        (data.FileRecord & {'exists': 1} &
         'dataset_name in ("_ibl_licks.times.npy", \
                           "_ibl_lickPiezo.raw.npy", \
                           "_ibl_lickPiezo.timestamps.npy")')

    def make(self, key):

        eID = (acquisition.Session & key).fetch1('session_uuid')

        lick_times, lick_piezo_raw, lick_piezo_timestamps = \
            ONE().load(eID, dataset_types=['_ibl_licks.times',
                                           '_ibl_lickPiezo.raw',
                                           '_ibl_lickPiezo.timestamps'])

        lick_sample_ids = lick_piezo_timestamps[:, 0]
        lick_piezo_timestamps = lick_piezo_timestamps[:, 1]

        key['lick_times'] = lick_times
        key['lick_piezo_raw'] = lick_piezo_raw
        key['lick_sample_ids'] = lick_sample_ids
        key['lick_piezo_timestamps'] = lick_piezo_timestamps
        key['lick_start_time'] = lick_piezo_timestamps[0]
        key['lick_end_time'] = lick_piezo_timestamps[-1]
        key['lick_sampling_rate'] = \
            1 / np.median(np.diff(lick_piezo_timestamps))

        self.insert1(key)

        logger.info('Populated a Lick tuple for \
            subject {subject_uuid} in session started at \
            {session_start_time}'.format(**key))


@schema
class CompleteTrialSession(dj.Computed):
    definition = """
    # sessions that are complete with trial information and thus may be ingested
    -> acquisition.Session
    ---
    stim_on_times_status: enum('Complete', 'Partial', 'Missing')
    rep_num_status: enum('Complete', 'Missing')
    included_status: enum('Complete', 'Missing')
    """

    required_datasets = ["_ibl_trials.feedback_times.npy",
                         "_ibl_trials.feedbackType.npy",
                         "_ibl_trials.intervals.npy", "_ibl_trials.choice.npy",
                         "_ibl_trials.response_times.npy",
                         "_ibl_trials.contrastLeft.npy",
                         "_ibl_trials.contrastRight.npy",
                         "_ibl_trials.probabilityLeft.npy"]

    def make(self, key):
        datasets = (data.FileRecord & key & 'repo_name LIKE "flatiron_%"' &
                    {'exists': 1}).fetch('dataset_name')
        is_complete = bool(np.all([req_ds in datasets
                                   for req_ds in self.required_datasets]))
        if is_complete is True:
            if '_ibl_trials.stimOn_times.npy' not in datasets:
                key['stim_on_times_status'] = 'Missing'
            else:
                eID = str((acquisition.Session & key).fetch1('session_uuid'))
                lab_name = (subject.SubjectLab & key).fetch1('lab_name')
                if lab_name == 'wittenlab':
                    stimOn_times = np.squeeze(ONE().load(
                            eID, dataset_types='_ibl_trials.stimOn_times',
                            clobber=True))
                else:
                    stimOn_times = ONE().load(
                        eID, dataset_types='_ibl_trials.stimOn_times')

                if np.all(np.isnan(stimOn_times)):
                    key['stim_on_times_status'] = 'Missing'
                elif np.any(np.isnan(stimOn_times)):
                    key['stim_on_times_status'] = 'Partial'
                else:
                    key['stim_on_times_status'] = 'Complete'

            if '_ibl_trials.repNum.npy' not in datasets:
                key['rep_num_status'] = 'Missing'
            else:
                key['rep_num_status'] = 'Complete'

            if '_ibl_trials.included.npy' not in datasets:
                key['included_status'] = 'Missing'
            else:
                key['included_status'] = 'Complete'

            self.insert1(key)


@schema
class TrialSet(dj.Imported):
    definition = """
    # information about behavioral trials
    -> acquisition.Session
    ---
    n_trials:                int      # total trial numbers in this set
    n_correct_trials=null:   int      # number of the correct trials
    trials_start_time:       float    # start time of the trial set (seconds)
    trials_end_time:         float    # end time of the trial set (seconds)
    """

    # Knowledge based hack to be formalized better later
    key_source = acquisition.Session & CompleteTrialSession

    def make(self, key):
        trial_key = key.copy()
        eID = str((acquisition.Session & key).fetch1('session_uuid'))

        trials_feedback_times, trials_feedback_types, trials_intervals, \
            trials_response_choice, trials_response_times, \
            trials_contrast_left, trials_contrast_right, trials_p_left = \
            ONE().load(eID, dataset_types=['_ibl_trials.feedback_times',
                                           '_ibl_trials.feedbackType',
                                           '_ibl_trials.intervals',
                                           '_ibl_trials.choice',
                                           '_ibl_trials.response_times',
                                           '_ibl_trials.contrastLeft',
                                           '_ibl_trials.contrastRight',
                                           '_ibl_trials.probabilityLeft'])

        stim_on_times_status, rep_num_status, included_status = \
            (CompleteTrialSession & key).fetch1(
                'stim_on_times_status', 'rep_num_status', 'included_status')

        lab_name = (subject.SubjectLab & key).fetch1('lab_name')
        if stim_on_times_status != 'Missing':
            if lab_name == 'wittenlab':
                trials_visual_stim_times = np.squeeze(ONE().load(
                    eID, dataset_types='_ibl_trials.stimOn_times',
                    clobber=True))
            else:
                trials_visual_stim_times = ONE().load(
                    eID, dataset_types='_ibl_trials.stimOn_times')

            if len(trials_visual_stim_times) == 1:
                trials_visual_stim_times = np.squeeze(trials_visual_stim_times)

        if rep_num_status != 'Missing':
            trials_rep_num = np.squeeze(ONE().load(
                eID, dataset_types='_ibl_trials.repNum'))

        if included_status != 'Missing':
            trials_included = np.squeeze(ONE().load(
                eID, dataset_types='_ibl_trials.included'))

        assert len(np.unique(np.array([len(trials_feedback_times),
                                       len(trials_feedback_types),
                                       len(trials_intervals),
                                       # len(trials_rep_num),
                                       len(trials_response_choice),
                                       len(trials_response_times),
                                       len(trials_contrast_left),
                                       len(trials_contrast_right),
                                       # len(trials_visual_stim_times),
                                       # len(trials_included),
                                       len(trials_p_left)
                                       ]))) == 1
        'Loaded trial files do not have the same length'

        key['n_trials'] = len(trials_response_choice)

        key_session = dict()
        key_session['model'] = 'actions.session'
        key_session['uuid'] = (acquisition.Session & key).fetch1(
            'session_uuid')

        n_correct_trials = grf(key_session, 'n_correct_trials')
        if n_correct_trials != 'None':
            key['n_correct_trials'] = n_correct_trials
        else:
            key['n_correct_trials'] = \
                sum((np.squeeze(trials_response_choice) == 1) &
                    (np.squeeze(trials_contrast_left) > 0)) \
                + sum((np.squeeze(trials_response_choice) == -1) &
                      (np.squeeze(trials_contrast_right) > 0))

        key['trials_start_time'] = trials_intervals[0, 0]
        key['trials_end_time'] = trials_intervals[-1, 1]

        self.insert1(key)

        for idx_trial in range(len(trials_response_choice)):

            if np.isnan(trials_contrast_left[idx_trial]):
                trial_stim_contrast_left = 0
            else:
                trial_stim_contrast_left = trials_contrast_left[idx_trial]

            if np.isnan(trials_contrast_right[idx_trial]):
                trial_stim_contrast_right = 0
            else:
                trial_stim_contrast_right = trials_contrast_right[idx_trial]

            if trials_response_choice[idx_trial] == -1:
                trial_response_choice = "CCW"
            elif trials_response_choice[idx_trial] == 0:
                trial_response_choice = "No Go"
            elif trials_response_choice[idx_trial] == 1:
                trial_response_choice = "CW"
            else:
                raise ValueError('Invalid reponse choice.')

            trial_key['trial_id'] = idx_trial + 1
            trial_key['trial_start_time'] = trials_intervals[idx_trial, 0]
            trial_key['trial_end_time'] = trials_intervals[idx_trial, 1]
            trial_key['trial_response_time'] = float(
                trials_response_times[idx_trial])
            trial_key['trial_response_choice'] = trial_response_choice

            if stim_on_times_status != 'Missing':
                trial_key['trial_stim_on_time'] = trials_visual_stim_times[
                    idx_trial]

            trial_key['trial_stim_contrast_left'] = float(
                trial_stim_contrast_left)
            trial_key['trial_stim_contrast_right'] = float(
                trial_stim_contrast_right)
            trial_key['trial_feedback_time'] = float(
                trials_feedback_times[idx_trial])
            trial_key['trial_feedback_type'] = int(
                trials_feedback_types[idx_trial])

            if rep_num_status != 'Missing':
                trial_key['trial_rep_num'] = int(trials_rep_num[idx_trial])

            trial_key['trial_stim_prob_left'] = float(trials_p_left[idx_trial])

            if included_status != 'Missing':
                trial_key['trial_included'] = bool(trials_included[idx_trial])

            self.Trial().insert1(trial_key)

        logger.info('Populated a TrialSet tuple, \
            all Trial tuples and Excluded Trial tuples for \
            subject {subject_uuid} in session started at \
            {session_start_time}'.format(**key))

    class Trial(dj.Part):
        # all times are in absolute seconds, rather than relative to trial onset
        definition = """
        -> master
        trial_id:               int           # trial identification number
        ---
        trial_start_time:           double         # beginning of quiescent period time (seconds)
        trial_end_time:             double         # end of iti (seconds)
        trial_response_time:        double         # Time of "response" in choiceworld (seconds). This is when one of the three possible choices is registered in software, will not be the same as when the mouse's movement to generate that response begins.
        trial_response_choice:      enum("CCW", "CW", "No Go")       # which choice was made in choiceworld
        trial_stim_on_time=null:    double         # Time of stimulus in choiceworld (seconds)
        trial_stim_contrast_left:   float	      # contrast of the stimulus on the left
        trial_stim_contrast_right:  float         # contrast of the stimulus on the right
        trial_feedback_time:        double         # Time of feedback delivery (reward or not) in choiceworld
        trial_feedback_type:        tinyint       # whether feedback is positive or negative in choiceworld (-1 for negative, +1 for positive)
        trial_rep_num=null:         int     	  # the repetition number of the trial, i.e. how many trials have been repeated on this side (counting from 1)
        trial_stim_prob_left:       float         # probability of the stimulus being present on left
        trial_included=null:        bool          # whether the trial should be included
        """

    class ExcludedTrial(dj.Part):
        definition = """
        -> master
        -> TrialSet.Trial
        """


@schema
class PassiveTrialSet(dj.Imported):
    definition = """
    -> acquisition.Session
    ---
    passive_trials_total_num : int
    passive_trials_start_time : float
    passive_trials_end_time : float
    """

    key_source = acquisition.Session & (data.FileRecord & {'exists': 1} &
                                        'dataset_name in \
                                        ("_ibl_passiveVisual.contrastLeft.npy", \
                                        "_ibl_passiveVisual.contrastRight.npy", \
                                        "_ibl_lickPiezo.timestamps.npy")')

    def make(self, key):

        passive_trial_key = key.copy()
        eID = str((acquisition.Session & key).fetch1('session_uuid'))

        passive_visual_stim_contrast_left, passive_visual_stim_contrast_right = \
            ONE().load(eID, dataset_types=['_ibl_passiveVisual.contrastLeft',
                                           '_ibl_passiveVisual.contrastRight',
                                           '_ibl_passiveVisual.times'])

        assert len(np.unique(np.array([len(passive_visual_stim_contrast_left),
                                       len(passive_visual_stim_contrast_right),
                                       len(passive_visual_stim_times)]))) == 1, \
            'Loaded passive visual files do not have the same length'

        key['passive_trials_total_num'] = len(passive_visual_stim_times)
        key['passive_trials_start_time'] = float(passive_visual_stim_times[0])
        key['passive_trials_end_time'] = float(passive_visual_stim_times[-1])

        self.insert1(key)

        for idx_trial in range(len(passive_visual_stim_times)):

            if np.isnan(passive_visual_stim_contrast_left[idx_trial]):
                passive_stim_contrast_left = 0
            else:
                passive_stim_contrast_left = \
                    passive_visual_stim_contrast_left[idx_trial]

            if np.isnan(passive_visual_stim_contrast_right[idx_trial]):
                passive_stim_contrast_right = 0
            else:
                passive_stim_contrast_right = \
                    passive_visual_stim_contrast_right[idx_trial]

            passive_trial_key['passive_trial_id'] = idx_trial + 1
            passive_trial_key['passive_trial_stim_on_time'] = float(
                passive_visual_stim_times[idx_trial])
            passive_trial_key['passive_trial_stim_contrast_left'] = float(
                passive_stim_contrast_left)
            passive_trial_key['passive_trial_stim_contrast_right'] = float(
                passive_stim_contrast_right)

            self.PassiveTrial().insert1(passive_trial_key)

        logger.info('Populated a PassiveTrialSet tuple, all Trial tuples and \
            Excluded Trial tuples for subject {subject_uuid} in \
                session started at {session_start_time}'.format(**key))

    class PassiveTrial(dj.Part):
        definition = """
        -> master
        passive_trial_id:           int         # trial identifier
        ---
        passive_trial_stim_on_time:          float	# Time of stimuli in choiceworld
        passive_trial_stim_contrast_left:    float 	# contrast of the stimulus on the left
        passive_trial_stim_contrast_right:   float   # contrast of the stimulus on the right
        """


@schema
class PassiveRecordings(dj.Imported):
    definition = """
    -> acquisition.Session
    ---
    passive_valve_click_times:      longblob      # Times of valve opening during passive trial presentation (seconds)
    passive_beep_times:             longblob      # Times of the beeps, equivilent to the go cue during the choice world task (seconds)
    passive_white_noise_times:      longblob      # Times of white noise bursts, equivilent to the negative feedback sound during the choice world task (seconds)
    """

    key_source = acquisition.Session & (data.FileRecord & 'repo_name LIKE "flatiron_%"' & {'exists': 1} & 'dataset_name in \
                    ("_ibl_passiveBeeps.times.npy", "_ibl_passiveValveClick.times.npy", "_ibl_passiveWhiteNoise.times.npy")')

    def make(self, key):

        eID = str((acquisition.Session & key).fetch1('session_uuid'))

        key['passive_beep_times'], key['passive_valve_click_times'], \
            key['passive_white_noise_times'] = \
            ONE().load(eID, dataset_types=['_ibl_passiveBeeps.times',
                                           '_ibl_passiveValveClick.times',
                                           '_ibl_passiveWhiteNoise.times'])

        self.insert1(key)



import datajoint as dj
from . import reference, acquisition

schema = dj.schema(dj.config.get('database.prefix', '') + 'ibl_dj_data')


@schema
class DataFormat(dj.Lookup):
    definition = """
    format_name:                    varchar(255)
    ---
    format_uuid:                    uuid
    file_extension='':              varchar(255)
    matlab_loader_function=null:    varchar(255)
    python_loader_function=null:    varchar(255)
    format_description=null:        varchar(255)
    """


@schema
class DataRepositoryType(dj.Lookup):
    definition = """
    repotype_name:  varchar(255)
    ---
    repotype_uuid:  uuid
    """


@schema
class DataRepository(dj.Lookup):
    definition = """
    repo_name:          varchar(255)
    ---
    -> DataRepositoryType
    repo_uuid:          uuid
    repo_timezone:      varchar(255)
    repo_hostname:      varchar(255)
    globus_endpoint_id: varchar(255)
    globus_path:        varchar(255)
    data_url=null:      varchar(255)
    globus_is_personal: boolean
    """


@schema
class ProjectRepository(dj.Manual):
    definition = """
    -> reference.Project
    -> DataRepository
    """


@schema
class DataSetType(dj.Lookup):
    definition = """
    dataset_type_name:              varchar(255)
    ---
    dataset_type_uuid:              uuid
    -> [nullable] reference.LabMember.proj(dataset_type_created_by='user_name')
    filename_pattern:               varchar(255)
    dataset_type_description=null:  varchar(1024)
    """


@schema
class DataSet(dj.Manual):
    definition = """
    -> acquisition.Session
    dataset_name:               varchar(255)
    ---
    dataset_uuid:               uuid
    -> reference.LabMember.proj(dataset_created_by='user_name')
    -> DataSetType
    -> DataFormat
    created_datetime:           datetime
    generating_software=null:   varchar(255)
    provenance_directory=null:  varchar(255)
    md5=null:                   varchar(255)
    file_size=null:             float
    """


@schema
class FileRecord(dj.Manual):
    definition = """
    -> DataSet
    -> DataRepository
    ---
    record_uuid:        uuid
    exists:             boolean
    relative_path:      varchar(255)
    """



import datajoint as dj
import uuid
from . import alyxraw, reference, subject
from . import get_raw_field as grf


schema = dj.schema(dj.config.get('database.prefix', '') +
                   'ibl_dj_ingest_action')


@schema
class ProcedureType(dj.Computed):
    definition = """
    (procedure_type_uuid) -> alyxraw.AlyxRaw
    ---
    procedure_type_name:                varchar(255)
    procedure_type_description=null:    varchar(1024)
    """

    key_source = (alyxraw.AlyxRaw & 'model="actions.proceduretype"').proj(
        procedure_type_uuid='uuid')

    def make(self, key):
        key_pt = key.copy()
        key['uuid'] = key['procedure_type_uuid']
        key_pt['procedure_type_name'] = grf(key, 'name')

        description = grf(key, 'description')
        if description != 'None':
            key_pt['procedure_type_description'] = description

        self.insert1(key_pt)


@schema
class Weighing(dj.Computed):
    # <class 'actions.models.Weighing'>
    definition = """
    (weigh_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:       uuid
    weighing_time:		datetime		# date time
    weight:	            float			# weight in grams
    weighing_user=null: varchar(255)
    """
    key_source = (alyxraw.AlyxRaw & 'model="actions.weighing"').proj(
        weigh_uuid='uuid')

    def make(self, key):
        key_weigh = key.copy()
        key['uuid'] = key['weigh_uuid']
        key_weigh['subject_uuid'] = uuid.UUID(grf(key, 'subject'))

        if not len(subject.Subject & key_weigh):
            print('subject {} is not in the Subject table'.format(
                key_weigh['subject_uuid']))
            return

        key_weigh['weighing_time'] = grf(key, 'date_time')

        weight = grf(key, 'weight')
        if weight != 'None':
            key_weigh['weight'] = float(weight)

        user_uuid = grf(key, 'user')
        if user_uuid != 'None':
            key_weigh['weighing_user'] = \
                (reference.LabMember &
                 dict(user_uuid=uuid.UUID(user_uuid))).fetch1('user_name')

        self.insert1(key_weigh)


@schema
class WaterType(dj.Computed):
    definition = """
    (watertype_uuid) -> alyxraw.AlyxRaw
    ---
    watertype_name:     varchar(255)
    """
    key_source = (alyxraw.AlyxRaw & 'model = "actions.watertype"').proj(
        watertype_uuid='uuid')

    def make(self, key):
        key_type = key.copy()
        key['uuid'] = key['watertype_uuid']

        key_type['watertype_name'] = grf(key, 'name')
        self.insert1(key_type)


@schema
class WaterAdministration(dj.Computed):
    # <class 'actions.models.WaterAdministration'>
    definition = """
    (wateradmin_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:                   uuid
    administration_user=null:       varchar(255)
    administration_time:	        datetime		# date time
    water_administered=null:		float			# water administered
    watertype_name:		            varchar(255)    # type of water
    adlib:                          boolean
    """
    key_source = (alyxraw.AlyxRaw &
                  'model = "actions.wateradministration"').proj(
        wateradmin_uuid='uuid')

    def make(self, key):
        key_wa = key.copy()
        key['uuid'] = key['wateradmin_uuid']

        key_wa['subject_uuid'] = uuid.UUID(grf(key, 'subject'))

        if not len(subject.Subject & key_wa):
            print('subject {} is not in the Subject table'.format(
                key_wa['subject_uuid']))
            return

        key_wa['administration_time'] = grf(key, 'date_time')
        wa = grf(key, 'water_administered')
        if wa != 'None':
            key_wa['water_administered'] = wa

        water_type = grf(key, 'water_type')
        key_wa['watertype_name'] = \
            (WaterType &
             dict(watertype_uuid=uuid.UUID(water_type))).fetch1(
                 'watertype_name')

        user_uuid = grf(key, 'user')
        if user_uuid != 'None':
            key_wa['administration_user'] = \
                (reference.LabMember &
                 dict(user_uuid=uuid.UUID(user_uuid))).fetch1('user_name')

        key_wa['adlib'] = grf(key, 'adlib') == 'True'

        self.insert1(key_wa)


@schema
class WaterRestriction(dj.Computed):
    # <class 'actions.models.WaterRestriction'>
    definition = """
    (restriction_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:               uuid
    restriction_start_time:     datetime	# start time
    restriction_end_time=null:  datetime	# end time
    restriction_narrative=null: varchar(1024)
    reference_weight:           float
    restriction_lab=null:       varchar(255)
    restriction_location=null:  varchar(255)
    """
    key_source = (alyxraw.AlyxRaw & 'model = "actions.waterrestriction"').proj(
        restriction_uuid='uuid')

    def make(self, key):
        key_res = key.copy()
        key['uuid'] = key['restriction_uuid']

        key_res['subject_uuid'] = uuid.UUID(grf(key, 'subject'))
        if not (subject.Subject & key_res):
            print('subject {} is not in the Subject table'.format(
                key_res['subject_uuid']))
            return

        key_res['restriction_start_time'] = grf(key, 'start_time')

        end_time = grf(key, 'end_time')
        if end_time != 'None':
            key_res['restriction_end_time'] = end_time

        narrative = grf(key, 'narrative')
        if narrative != 'None':
            key_res['restriction_narrative'] = narrative

        location_uuid = grf(key, 'location')
        if location_uuid != 'None':
            key_res['restriction_lab'], key_res['restriction_location'] = \
                (reference.LabLocation &
                 dict(location_uuid=uuid.UUID(location_uuid))).fetch1(
                     'lab_name', 'location_name')

        key_res['reference_weight'] = grf(key, 'reference_weight')

        self.insert1(key_res)


@schema
class WaterRestrictionUser(dj.Manual):
    definition = """
    subject_uuid:           uuid
    restriction_start_time: datetime
    user_name:              varchar(255)
    """


@schema
class WaterRestrictionProcedure(dj.Manual):
    definition = """
    subject_uuid:           uuid
    restriction_start_time: varchar(255)
    procedure_type_name:    varchar(255)
    """


@schema
class Surgery(dj.Computed):
    # <class 'actions.models.Surgery'>
    definition = """
    (surgery_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:               uuid
    surgery_lab=null:           varchar(255)    # lab for sugery
    surgery_location=null:      varchar(255)    # foreign key inherited from reference.LabLocation
    surgery_start_time:	        datetime        # surgery start time
    surgery_end_time=null:	    datetime        # surgery end time
    surgery_outcome_type:		enum('None', 'a', 'n', 'r')	    # outcome type
    surgery_narrative=null:     varchar(2048)    	# narrative
    """
    key_source = (alyxraw.AlyxRaw & 'model = "actions.surgery"').proj(
        surgery_uuid='uuid')

    def make(self, key):
        key_surgery = key.copy()
        key['uuid'] = key['surgery_uuid']

        key_surgery['subject_uuid'] = uuid.UUID(grf(key, 'subject'))
        if not len(subject.Subject & key_surgery):
            print('Subject {} is not in the table subject.Subject'.format(
                key_surgery['surgery_uuid']
            ))

        key_surgery['surgery_start_time'] = grf(key, 'start_time')

        end_time = grf(key, 'end_time')
        if end_time != 'None':
            key_surgery['surgery_end_time'] = end_time

        key_surgery['surgery_outcome_type'] = grf(key, 'outcome_type')

        narrative = grf(key, 'narrative')
        if narrative != 'None':
            key_surgery['surgery_narrative'] = narrative

        location_uuid = grf(key, 'location')
        if location_uuid != 'None':
            key_surgery['surgery_lab'], key_surgery['surgery_location'] = \
                (reference.LabLocation &
                 dict(location_uuid=uuid.UUID(location_uuid))).fetch1(
                     'lab_name', 'location_name')

        self.insert1(key_surgery)


@schema
class SurgeryUser(dj.Manual):
    definition = """
    subject_uuid:       uuid
    surgery_start_time: datetime
    user_name:          varchar(255)
    """


@schema
class SurgeryProcedure(dj.Manual):
    definition = """
    subject_uuid:           uuid
    surgery_start_time:     datetime
    procedure_type_name:    varchar(255)
    """


@schema
class VirusInjection(dj.Computed):
    # <class 'actions.models.VirusInjection'>
    definition = """
    (virus_injection_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:           uuid
    injection_time:		    datetime        	# injection time
    injection_volume:		float   		    # injection volume
    rate_of_injection:		float               # rate of injection
    injection_type:		    varchar(255)    	# injection type
    """
    key_source = alyxraw.AlyxRaw & 'model = "actions.virusinjection"'
    # data missing


@schema
class OtherAction(dj.Computed):
    # <class 'actions.models.OtherAction'>
    definition = """
    (other_action_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:               uuid
    other_action_start_time:    datetime	    # start time
    other_action_end_time=null: datetime	    # end time
    other_action_lab=null:      varchar(255)
    other_action_location=null: varchar(255)    # refer to reference.Location
    """
    key_source = (alyxraw.AlyxRaw & 'model = "actions.otheraction"').proj(
        other_action_uuid='uuid')

    def make(self, key):
        key_other = key.copy()
        key['uuid'] = key['other_action_uuid']
        key_other['subject_uuid'] = uuid.UUID(grf(key, 'subject'))

        if not len(subject.Subject & key_other):
            print('subject {} is not in the table subject.Subject'.format(
                key_other['subject_uuid']))
            return

        key_other['other_action_start_time'] = grf(key, 'start_time')

        end_time = grf(key, 'end_time')
        if end_time != 'None':
            key_other['other_action_end_time'] = end_time

        location_uuid = grf(key, 'location')
        if location_uuid != 'None':
            key_other['other_action_lab'], key_other['other_action_location'] = \
                (reference.LabLocation &
                 dict(location_uuid=uuid.UUID(location_uuid))).fetch1(
                     'lab_name', 'location_name')

        self.insert1(key_other)


@schema
class OtherActionUser(dj.Manual):
    # <class 'actions.models.OtherAction'>
    definition = """
    subject_uuid:               uuid
    other_action_start_time:    datetime	    # start time
    user_name:        varchar(255)
    """


@schema
class OtherActionProcedure(dj.Manual):
    # <class 'actions.models.OtherAction'>
    definition = """
    subject_uuid:               uuid
    other_action_start_time:    datetime	    # start time
    procedure_type_name:        varchar(255)
    """



import datajoint as dj
import json
import uuid

from . import alyxraw, reference, subject, action
from . import get_raw_field as grf

schema = dj.schema(dj.config.get('database.prefix', '') +
                   'ibl_dj_ingest_acquisition')


@schema
class Session(dj.Computed):
    definition = """
    (session_uuid) -> alyxraw.AlyxRaw
    ---
    session_number=null:        int
    subject_uuid:               uuid
    project_name=null:          varchar(255)
    session_start_time:         datetime
    session_end_time=null:      datetime
    session_lab=null:           varchar(255)
    session_location=null:      varchar(255)
    session_type=null:          varchar(255)
    session_narrative=null:     varchar(1024)
    task_protocol=null:         varchar(255)
    """
    key_source = (alyxraw.AlyxRaw & 'model="actions.session"').proj(
        session_uuid='uuid')

    def make(self, key):
        key_session = key.copy()
        key['uuid'] = key['session_uuid']
        key_session['subject_uuid'] = uuid.UUID(grf(key, 'subject'))

        if not len(subject.Subject & key_session):
            print('Subject {} is not in the table subject.Subject'.format(
                key_session['subject_uuid']))
            return

        session_number = grf(key, 'number')
        if session_number != 'None':
            key_session['session_number'] = session_number

        proj_uuid = grf(key, 'project')
        if proj_uuid != 'None':
            key_session['project_name'] = \
                (reference.Project &
                 dict(project_uuid=uuid.UUID(proj_uuid))).fetch1(
                    'project_name')

        key_session['session_start_time'] = grf(key, 'start_time')

        end_time = grf(key, 'end_time')
        if end_time != 'None':
            key_session['session_end_time'] = end_time

        location_uuid = grf(key, 'location')
        if location_uuid != 'None':
            key_session['session_lab'], key_session['session_location'] = \
                (reference.LabLocation &
                 dict(location_uuid=uuid.UUID(location_uuid))).fetch1(
                     'lab_name', 'location_name')

        session_type = grf(key, 'type')
        if session_type != 'None':
            key_session['session_type'] = session_type

        narrative = grf(key, 'narrative')
        if narrative != 'None' and narrative != "":
            key_session['session_narrative'] = narrative

        protocol = grf(key, 'task_protocol')
        if protocol != 'None':
            key_session['task_protocol'] = protocol

        self.insert1(key_session)


@schema
class ChildSession(dj.Manual):
    definition = """
    subject_uuid:               uuid
    session_start_time:         datetime
    ---
    parent_session_start_time:  datetime
    """


@schema
class SessionUser(dj.Manual):
    definition = """
    subject_uuid:           uuid
    session_start_time:     datetime
    user_name:              varchar(255)
    """


@schema
class SessionProcedure(dj.Manual):
    definition = """
    subject_uuid:           uuid
    session_start_time:     datetime
    procedure_type_name:    varchar(255)
    """


@schema
class WaterAdministrationSession(dj.Manual):
    definition = """
    subject_uuid:           uuid
    administration_time:    datetime
    ---
    session_start_time:     datetime
    """



import datajoint as dj
import json
import uuid

from . import alyxraw, reference, acquisition
from . import get_raw_field as grf

schema = dj.schema(dj.config.get('database.prefix', '') +
                   'ibl_dj_ingest_data')


@schema
class DataFormat(dj.Computed):
    definition = """
    (format_uuid) -> alyxraw.AlyxRaw
    ---
    format_name:                    varchar(255)
    file_extension='':              varchar(255)
    matlab_loader_function=null:    varchar(255)
    python_loader_function=null:    varchar(255)
    format_description=null:        varchar(255)
    """
    key_source = (alyxraw.AlyxRaw & 'model="data.dataformat"').proj(
        format_uuid='uuid')

    def make(self, key):
        key_format = key.copy()
        key['uuid'] = key['format_uuid']

        key_format['format_name'] = grf(key, 'name')

        file_extension = grf(key, 'file_extension')
        if file_extension != 'None':
            key_format['file_extension'] = file_extension

        matlab_loader_function = grf(key, 'matlab_loader_function')
        if matlab_loader_function != 'None':
            key_format['matlab_loader_function'] = matlab_loader_function

        python_loader_function = grf(key, 'python_loader_function')
        if python_loader_function != 'None':
            key_format['python_loader_function'] = python_loader_function

        format_description = grf(key, 'description')
        if format_description != 'None':
            key_format['format_description'] = format_description

        self.insert1(key_format)


@schema
class DataRepositoryType(dj.Computed):
    definition = """
    (repotype_uuid) -> alyxraw.AlyxRaw
    ---
    repotype_name: varchar(255)
    """
    key_source = (alyxraw.AlyxRaw & 'model="data.datarepositorytype"').proj(
        repotype_uuid='uuid')

    def make(self, key):
        key_repotype = key.copy()
        key['uuid'] = key['repotype_uuid']
        key_repotype['repotype_name'] = grf(key, 'name')
        self.insert1(key_repotype)


@schema
class DataRepository(dj.Computed):
    definition = """
    (repo_uuid) -> alyxraw.AlyxRaw
    ---
    repo_name:          varchar(255)
    repotype_name:      varchar(255)
    repo_timezone:     varchar(255)
    repo_hostname:      varchar(255)
    globus_endpoint_id: varchar(255)
    globus_path:        varchar(255)
    data_url=null:      varchar(255)
    globus_is_personal: boolean
    """
    key_source = (alyxraw.AlyxRaw & 'model="data.datarepository"').proj(
        repo_uuid='uuid')

    def make(self, key):
        key_repo = key.copy()
        key['uuid'] = key['repo_uuid']

        key_repo['repo_name'] = grf(key, 'name')

        repotype = grf(key, 'repository_type')
        key_repo['repotype_name'] = \
            (DataRepositoryType &
                dict(repotype_uuid=uuid.UUID(repotype))).fetch1(
                    'repotype_name')
        key_repo['repo_timezone'] = grf(key, 'timezone')
        key_repo['repo_hostname'] = grf(key, 'hostname')
        key_repo['globus_endpoint_id'] = grf(key, 'globus_endpoint_id')
        key_repo['globus_path'] = grf(key, 'globus_path')

        is_personal = grf(key, 'globus_is_personal')
        key_repo['globus_is_personal'] = True if is_personal else False

        url = grf(key, 'data_url')
        if url != 'None':
            key_repo['data_url'] = url

        self.insert1(key_repo)


@schema
class ProjectRepository(dj.Manual):
    definition = """
    project_name:       varchar(255)
    repo_name:          varchar(255)
    """


@schema
class DataSetType(dj.Computed):
    definition = """
    (dataset_type_uuid) -> alyxraw.AlyxRaw
    ---
    dataset_type_name:              varchar(255)
    dataset_type_created_by=null:   varchar(255)
    filename_pattern:               varchar(255)
    dataset_type_description=null:  varchar(1024)
    """
    key_source = (alyxraw.AlyxRaw & 'model="data.datasettype"').proj(
        dataset_type_uuid='uuid')

    def make(self, key):
        key_dst = key.copy()
        key['uuid'] = key['dataset_type_uuid']

        key_dst['dataset_type_name'] = grf(key, 'name')

        user_uuid = grf(key, 'created_by')
        if user_uuid != 'None':
            key_dst['dataset_type_created_by'] = \
                (reference.LabMember &
                 dict(user_uuid=uuid.UUID(user_uuid))).fetch1('user_name')

        key_dst['filename_pattern'] = grf(key, 'filename_pattern')
        key_dst['dataset_type_description'] = grf(key, 'description')

        self.insert1(key_dst)


@schema
class DataSet(dj.Computed):
    definition = """
    (dataset_uuid) -> alyxraw.AlyxRaw
    ---
    subject_uuid:               uuid
    session_start_time:         datetime
    dataset_created_by:         varchar(255)
    dataset_name:               varchar(255)
    dataset_type_name:          varchar(255)
    format_name:                varchar(255)
    created_datetime:           datetime
    generating_software=null:   varchar(255)
    provenance_directory=null:  varchar(255)
    md5=null:                   varchar(255)
    file_size=null:             float
    """
    key_source = (alyxraw.AlyxRaw & 'model="data.dataset"').proj(
        dataset_uuid="uuid")

    def make(self, key):
        key_ds = key.copy()
        key['uuid'] = key['dataset_uuid']

        session = grf(key, 'session')
        if not len(acquisition.Session &
                   dict(session_uuid=uuid.UUID(session))):
            print('Session {} is not in the table acquisition.Session'.format(
                session))
            return

        key_ds['subject_uuid'], key_ds['session_start_time'] = \
            (acquisition.Session &
             dict(session_uuid=uuid.UUID(session))).fetch1(
                'subject_uuid', 'session_start_time')

        key_ds['dataset_name'] = grf(key, 'name')

        dt = grf(key, 'dataset_type')
        key_ds['dataset_type_name'] = \
            (DataSetType & dict(dataset_type_uuid=uuid.UUID(dt))).fetch1(
                'dataset_type_name')

        user = grf(key, 'created_by')
        key_ds['dataset_created_by'] = \
            (reference.LabMember & dict(user_uuid=uuid.UUID(user))).fetch1(
                'user_name')

        format = grf(key, 'data_format')
        key_ds['format_name'] = \
            (DataFormat & dict(format_uuid=uuid.UUID(format))).fetch1(
                'format_name')

        key_ds['created_datetime'] = grf(key, 'created_datetime')

        software = grf(key, 'generating_software')
        if software != 'None':
            key_ds['generating_software'] = software

        directory = grf(key, 'provenance_directory')
        if directory != 'None':
            key_ds['provenance_directory'] = directory

        md5 = grf(key, 'md5')
        if md5 != 'None':
            key_ds['md5'] = md5

        file_size = grf(key, 'file_size')
        if file_size != 'None':
            key_ds['file_size'] = file_size

        self.insert1(key_ds)


@schema
class FileRecord(dj.Computed):
    definition = """
    (record_uuid) -> alyxraw.AlyxRaw
    ---
    exists:                     boolean
    subject_uuid:               uuid
    session_start_time:         datetime
    dataset_name:               varchar(255)
    repo_name:                  varchar(255)
    relative_path:              varchar(255)
    """
    records = alyxraw.AlyxRaw & 'model="data.filerecord"'
    repos = (DataRepository & 'repo_name LIKE "flatiron%"').fetch(
        'repo_uuid')
    records_flatiron = alyxraw.AlyxRaw.Field & records & \
        'fname = "data_repository"' & [{'fvalue': str(repo)} for repo in repos]
    record_exists = alyxraw.AlyxRaw.Field & records & \
        'fname = "exists"' & 'fvalue="True"'
    key_source = (alyxraw.AlyxRaw & record_exists & records_flatiron).proj(
        record_uuid='uuid')

    def make(self, key):
        key_fr = key.copy()
        key['uuid'] = key['record_uuid']
        key_fr['exists'] = True

        dataset = grf(key, 'dataset')
        if not len(DataSet & dict(dataset_uuid=uuid.UUID(dataset))):
            print('Dataset {} is not in the table data.DataSet')
            return

        key_fr['subject_uuid'], key_fr['session_start_time'], \
            key_fr['dataset_name'] = \
            (DataSet & dict(dataset_uuid=uuid.UUID(dataset))).fetch1(
                'subject_uuid', 'session_start_time', 'dataset_name')

        repo = grf(key, 'data_repository')
        key_fr['repo_name'] = \
            (DataRepository & dict(repo_uuid=uuid.UUID(repo))).fetch1(
                'repo_name')

        key_fr['relative_path'] = grf(key, 'relative_path')
        self.insert1(key_fr)



#!/usr/bin/env python
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

setup(
    name='ibl_pipeline',
    version='0.0.dev',
    description='Datajoint schemas for IBL',
    author='Vathes',
    author_email='support@vathes.com',
    packages=find_packages(exclude=[]),
    install_requires=['datajoint>=0.12.dev3'],
    scripts=['scripts/ibl-shell.py'],
)



"""
The following is provided as a convenience utility to be able to add/drop non primary key attributes
to an existing DataJoint table. Both function should work as is - you can simply copy and paste them and
start using it without worrying about importing dependencies.

WARNING! These functions are NOT officially supported by DataJoint and make use of DataJoint internal logic
that is considered to be NOT part of the public API. These functions may cease to work or worse yet result
in unexpected results without any prior notice, and therefore should be used with atmost care.

We are working to add such "ALTER" method to DataJoint Python officially, and you can track the discussion/progress
at https://github.com/datajoint/datajoint-python/issues/110.
"""

def add_column(table, name, dtype, default_value=None, comment=None):
    """
    A (hacky) convenience function to add a new column into an existing table.
    
    Args:
        table (DataJoint table class instance): table to add new column (attribute) to
        name (str): name of the new column
        dtype (str): data type of the new column
        default_value (str, optional): default value for the new column. If 'null' or None, then the attribute 
            is considered non-required. Defaults to None.
        comment (str, optional): comment for the new column
    """
    full_table_name = table.full_table_name
    if default_value is None or default_value.strip().lower() == 'null':
        query = 'ALTER TABLE {} ADD {} {}'.format(full_table_name, name, dtype)
    else:
        query = 'ALTER TABLE {} ADD {} {} NOT NULL DEFAULT {}'.format(full_table_name, name, dtype, repr(default_value))
    if comment is not None:
        query += ' COMMENT "{}"'.format(comment)
    table.connection.query(query)
    print('Be sure to add following entry to your table definition')
    definition = '{}={}: {}'.format(name, repr(default_value), dtype)
    if comment is not None:
        definition += ' # {}'.format(comment)
    print(definition)
    table.__class__._heading = None


def alter_column(table, name, dtype, default_value=None, comment=None):
    """
    A (hacky) convenience function to alter an existing column's definition - use with ultra caution!
    
    Args:
        table (DataJoint table class instance): table in which to modify a column's definition
        name (str): name of the column to be modified
        dtype (str): (new) data type of the column
        default_value (str, optional): (new) default value for the column. If 'null', then the attribute 
            is considered non-required. If None (default), the attribute is required.
        comment (str, optional): (new) comment for the new column
    """
    full_table_name = table.full_table_name
    if default_value is None:
        query = 'ALTER TABLE {} MODIFY {} {} NOT NULL'.format(full_table_name, name, dtype)
    elif default_value.strip().lower() == 'null':
        query = 'ALTER TABLE {} MODIFY {} {}'.format(full_table_name, name, dtype)
    else:
        query = 'ALTER TABLE {} MODIFY {} {} NOT NULL DEFAULT {}'.format(full_table_name, name, dtype, repr(default_value))
    if comment is not None:
        query += ' COMMENT "{}"'.format(comment)
    table.connection.query(query)
    print('Be sure to alter the column definition as follows')
    definition = '{}={}: {}'.format(name, repr(default_value), dtype)
    if comment is not None:
        definition += ' # {}'.format(comment)
    print(definition)
    table.__class__._heading = None
    
def drop_column(table, name):
    """
    Convenience function to drop specified column with name from the table.
    A primary key attribute may not be dropped. If name specifies a primary
    key attribute, this function raises a `ValueError`.
    
    Args:
        table (DataJoint table class instance): table to drop an attribute from
        name (str): name of the attribute to be dropped. Cannot be a primary key attribute.
    """
    from datajoint.utils import user_choice
    if name in table.primary_key:
        raise ValueError('You cannot drop a primary key attribute')
    if name not in table.heading.attributes:
        raise AttributeError('column {} not found in the table'.format(name))
    choice = user_choice('You are about to drop the column "{}". Proceed?'.format(name),
                         default='no')
    if choice == 'yes':
        query = 'ALTER TABLE {} DROP COLUMN {}'.format(table.full_table_name, name)
        table.connection.query(query)
        table.__class__._heading = None
        print('Dropeed column "{}"'.format(name))
    else:
        print('Aborting column drop')